# Allow classes to use self-referencing Type hints in Python 3.7.
from __future__ import annotations

from typing import Any, Dict, List
from uuid import uuid4

from deltacat.storage.model.sort_key import SortKey
from deltacat.storage import Locator, PartitionLocator
from deltacat.utils.common import sha1_hexdigest


class PrimaryKeyIndexLocator(Locator, dict):
    @staticmethod
    def of(primary_key_index_meta: PrimaryKeyIndexMeta) -> PrimaryKeyIndexLocator:
        """
        Creates a Primary Key Index Locator from the given Primary Key
        Index Metadata. A Primary Key Index Locator consists of a Primary Key
        Index Root Path common to all versions. This Primary Key Index Root
        Path is generated deterministically from Primary Key Index Metadata.
        """
        pki_root_path = PrimaryKeyIndexLocator._root_path(
            primary_key_index_meta.compacted_partition_locator,
            primary_key_index_meta.primary_keys,
            primary_key_index_meta.sort_keys,
            primary_key_index_meta.primary_key_index_algorithm_version,
        )
        pkil = PrimaryKeyIndexLocator()
        pkil["primaryKeyIndexMeta"] = primary_key_index_meta
        pkil["primaryKeyIndexRootPath"] = pki_root_path
        return pkil

    @staticmethod
    def _root_path(
        compacted_partition_locator: PartitionLocator,
        primary_keys: List[str],
        sort_keys: List[SortKey],
        primary_key_index_algorithm_version: str,
    ) -> str:
        pl_hexdigest = compacted_partition_locator.hexdigest()
        pki_version_str = (
            f"{pl_hexdigest}|{primary_keys}|{sort_keys}|"
            f"{primary_key_index_algorithm_version}"
        )
        return sha1_hexdigest(pki_version_str.encode("utf-8"))

    @property
    def primary_key_index_meta(self) -> PrimaryKeyIndexMeta:
        val: Dict[str, Any] = self.get("primaryKeyIndexMeta")
        if val is not None and not isinstance(val, PrimaryKeyIndexMeta):
            self["primaryKeyIndexMeta"] = val = PrimaryKeyIndexMeta(val)
        return val

    @property
    def primary_key_index_root_path(self) -> str:
        """
        Gets the root path for all sibling versions of the given Primary Key
        Index Locator.
        """
        return self["primaryKeyIndexRootPath"]

    def get_primary_key_index_s3_url_base(self, s3_bucket: str) -> str:
        """
        Gets the base S3 URL for all sibling versions of the given Primary Key
        Index Locator.
        """
        pki_root_path = self.primary_key_index_root_path
        return f"s3://{s3_bucket}/{pki_root_path}"

    def canonical_string(self) -> str:
        """
        Returns a unique string for the given locator that can be used
        for equality checks (i.e. two locators are equal if they have
        the same canonical string).
        """
        return self.primary_key_index_root_path


class PrimaryKeyIndexMeta(dict):
    @staticmethod
    def of(
        compacted_partition_locator: PartitionLocator,
        primary_keys: List[str],
        sort_keys: List[SortKey],
        primary_key_index_algo_version: str,
    ) -> PrimaryKeyIndexMeta:
        """
        Creates Primary Key Index Metadata from the given compacted
        Partition Locator, primary keys, sort keys, and primary key index
        algorithm version.
        """
        pkim = PrimaryKeyIndexMeta()
        pkim["compactedPartitionLocator"] = compacted_partition_locator
        pkim["primaryKeys"] = primary_keys
        pkim["sortKeys"] = sort_keys
        pkim["primaryKeyIndexAlgorithmVersion"] = primary_key_index_algo_version
        return pkim

    @property
    def compacted_partition_locator(self) -> PartitionLocator:
        val: Dict[str, Any] = self.get("compactedPartitionLocator")
        if val is not None and not isinstance(val, PartitionLocator):
            self["compactedPartitionLocator"] = val = PartitionLocator(val)
        return val

    @property
    def primary_keys(self) -> List[str]:
        return self["primaryKeys"]

    @property
    def sort_keys(self) -> List[SortKey]:
        return self["sortKeys"]

    @property
    def primary_key_index_algorithm_version(self) -> str:
        return self["primaryKeyIndexAlgorithmVersion"]


class PrimaryKeyIndexVersionLocator(Locator, dict):
    @staticmethod
    def of(
        primary_key_index_version_meta: PrimaryKeyIndexVersionMeta,
        pki_version_root_path: str,
    ) -> PrimaryKeyIndexVersionLocator:
        """
        Creates a primary key index version locator from the given primary key
        index version metadata and version root path. Note that, while this is
        useful for constructing a locator pointing to an existing primary key
        index version, the `generate` function should be used to create unique
        locators for all new primary key index versions.
        """
        pkivl = PrimaryKeyIndexVersionLocator()
        pkivl["primaryKeyIndexVersionMeta"] = primary_key_index_version_meta
        pkivl["primaryKeyIndexVersionRootPath"] = pki_version_root_path
        return pkivl

    @staticmethod
    def generate(
        pki_version_meta: PrimaryKeyIndexVersionMeta,
    ) -> PrimaryKeyIndexVersionLocator:
        """
        Creates a new primary key index version locator from the given primary
        key index version metadata. A primary key index version locator
        consists of a primary key index root path common to all versions, and
        a primary key index version root path common to all hash buckets for
        that version. The primary key index version root path is generated
        non-deterministically from the number of hash buckets used for this
        version and a new UUID. The primary key index root path is generated
        deterministically from the compacted partition locator, primary keys,
        sort keys, and primary key index algorithm version.
        """
        pki_version_root_path = (
            PrimaryKeyIndexVersionLocator._generate_version_root_path(
                PrimaryKeyIndexVersionLocator._pki_root_path(pki_version_meta),
                pki_version_meta.hash_bucket_count,
            )
        )
        pkivl = PrimaryKeyIndexVersionLocator()
        pkivl["primaryKeyIndexVersionMeta"] = pki_version_meta
        pkivl["primaryKeyIndexVersionRootPath"] = pki_version_root_path
        return pkivl

    @staticmethod
    def _pki_root_path(pki_version_meta: PrimaryKeyIndexVersionMeta) -> str:
        pki_meta = pki_version_meta.primary_key_index_meta
        pki_locator = PrimaryKeyIndexLocator.of(pki_meta)
        return pki_locator.primary_key_index_root_path

    @staticmethod
    def _generate_version_root_path(pki_root_path: str, hash_bucket_count: int) -> str:
        return f"{pki_root_path}/{hash_bucket_count}/{str(uuid4())}"

    @property
    def primary_key_index_version_meta(self) -> PrimaryKeyIndexVersionMeta:
        val: Dict[str, Any] = self.get("primaryKeyIndexVersionMeta")
        if val is not None and not isinstance(val, PrimaryKeyIndexVersionMeta):
            self["primaryKeyIndexVersionMeta"] = val = PrimaryKeyIndexVersionMeta(val)
        return val

    @property
    def primary_key_index_root_path(self) -> str:
        """
        Gets the root path for all sibling versions of the given primary key
        index version locator.
        """
        return PrimaryKeyIndexVersionLocator._pki_root_path(
            self.primary_key_index_version_meta
        )

    @property
    def primary_key_index_version_root_path(self) -> str:
        """
        Gets the root path for the primary key index version associated with
        the given primary key index version locator.
        """
        return self["primaryKeyIndexVersionRootPath"]

    def get_primary_key_index_version_s3_url_base(self, s3_bucket: str) -> str:
        """
        Gets the base S3 URL for the primary key index version associated with
        the given primary key index version locator.
        """
        pkiv_root_path = self.primary_key_index_version_root_path
        return f"s3://{s3_bucket}/{pkiv_root_path}"

    def get_pkiv_hb_index_root_path(self, hb_index: int) -> str:
        """
        Gets the root path of a single hash bucket of the given primary key
        index version locator.
        """
        pkiv_root_path = self.primary_key_index_version_root_path
        return f"{pkiv_root_path}/{hb_index}"

    def get_pkiv_hb_index_s3_url_base(
        self, s3_bucket: str, hash_bucket_index: int
    ) -> str:
        """
        Gets the base S3 URL of a single hash bucket of the given primary key
        index version locator.
        """
        hbi_root_path = self.get_pkiv_hb_index_root_path(hash_bucket_index)
        return f"s3://{s3_bucket}/{hbi_root_path}"

    def get_pkiv_hb_index_manifest_s3_url(
        self, s3_bucket: str, hash_bucket_index: int
    ) -> str:
        """
        Gets the S3 URL of the manifest for a single primary key index version
        hash bucket.
        """
        pkiv_hb_index_s3_url_base = self.get_pkiv_hb_index_s3_url_base(
            s3_bucket,
            hash_bucket_index,
        )
        return f"{pkiv_hb_index_s3_url_base}.json"

    def canonical_string(self) -> str:
        """
        Returns a unique string for the given locator that can be used
        for equality checks (i.e. two locators are equal if they have
        the same canonical string).
        """
        return self.primary_key_index_version_root_path


class PrimaryKeyIndexVersionMeta(dict):
    @staticmethod
    def of(
        primary_key_index_meta: PrimaryKeyIndexMeta, hash_bucket_count: int
    ) -> PrimaryKeyIndexVersionMeta:
        """
        Creates Primary Key Index Version Metadata from the given Primary Key
        Index Metadata and hash bucket count.
        """
        pkivm = PrimaryKeyIndexVersionMeta()
        pkivm["primaryKeyIndexMeta"] = primary_key_index_meta
        pkivm["hashBucketCount"] = hash_bucket_count
        return pkivm

    @property
    def primary_key_index_meta(self) -> PrimaryKeyIndexMeta:
        val: Dict[str, Any] = self.get("primaryKeyIndexMeta")
        if val is not None and not isinstance(val, PrimaryKeyIndexMeta):
            self["primaryKeyIndexMeta"] = val = PrimaryKeyIndexMeta(val)
        return val

    @property
    def hash_bucket_count(self) -> int:
        return self["hashBucketCount"]



# Allow classes to use self-referencing Type hints in Python 3.7.
from __future__ import annotations

from typing import Any, Dict, Optional

from deltacat.storage.model.locator import Locator


class Namespace(dict):
    @staticmethod
    def of(
        locator: Optional[NamespaceLocator], permissions: Optional[Dict[str, Any]]
    ) -> Namespace:
        namespace = Namespace()
        namespace.locator = locator
        namespace.permissions = permissions
        return namespace

    @property
    def locator(self) -> Optional[NamespaceLocator]:
        val: Dict[str, Any] = self.get("namespaceLocator")
        if val is not None and not isinstance(val, NamespaceLocator):
            self.locator = val = NamespaceLocator(val)
        return val

    @locator.setter
    def locator(self, namespace_locator: Optional[NamespaceLocator]) -> None:
        self["namespaceLocator"] = namespace_locator

    @property
    def namespace(self) -> Optional[str]:
        namespace_locator = self.locator
        if namespace_locator:
            return namespace_locator.namespace
        return None

    @property
    def permissions(self) -> Optional[Dict[str, Any]]:
        return self.get("permissions")

    @permissions.setter
    def permissions(self, permissions: Optional[Dict[str, Any]]) -> None:
        self["permissions"] = permissions


class NamespaceLocator(Locator, dict):
    @staticmethod
    def of(namespace: Optional[str]) -> NamespaceLocator:
        namespace_locator = NamespaceLocator()
        namespace_locator.namespace = namespace
        return namespace_locator

    @property
    def namespace(self) -> Optional[str]:
        return self.get("namespace")

    @namespace.setter
    def namespace(self, namespace: Optional[str]) -> None:
        self["namespace"] = namespace

    def canonical_string(self) -> str:
        """
        Returns a unique string for the given locator that can be used
        for equality checks (i.e. two locators are equal if they have
        the same canonical string).
        """
        return self.namespace



from unittest import TestCase
from deltacat.utils.pyarrow import (
    s3_partial_parquet_file_to_table,
    pyarrow_read_csv,
    content_type_to_reader_kwargs,
    _add_column_kwargs,
    s3_file_to_table,
    ReadKwargsProviderPyArrowSchemaOverride,
    RAISE_ON_EMPTY_CSV_KWARG,
)
from deltacat.types.media import ContentEncoding, ContentType
from deltacat.types.partial_download import PartialParquetParameters
from pyarrow.parquet import ParquetFile
import pyarrow as pa

PARQUET_FILE_PATH = "deltacat/tests/utils/data/test_file.parquet"
EMPTY_UTSV_PATH = "deltacat/tests/utils/data/empty.csv"
NON_EMPTY_VALID_UTSV_PATH = "deltacat/tests/utils/data/non_empty_valid.csv"
GZIP_COMPRESSED_FILE_UTSV_PATH = "deltacat/tests/utils/data/non_empty_compressed.gz"
BZ2_COMPRESSED_FILE_UTSV_PATH = "deltacat/tests/utils/data/non_empty_compressed.bz2"


class TestS3PartialParquetFileToTable(TestCase):
    def test_s3_partial_parquet_file_to_table_sanity(self):

        pq_file = ParquetFile(PARQUET_FILE_PATH)
        partial_parquet_params = PartialParquetParameters.of(
            pq_metadata=pq_file.metadata
        )

        self.assertEqual(
            partial_parquet_params.num_row_groups, 2, "test_file.parquet has changed."
        )

        # only first row group to be downloaded
        partial_parquet_params.row_groups_to_download.pop()

        result = s3_partial_parquet_file_to_table(
            PARQUET_FILE_PATH,
            include_columns=["n_legs"],
            content_encoding=ContentEncoding.IDENTITY.value,
            content_type=ContentType.PARQUET.value,
            partial_file_download_params=partial_parquet_params,
        )

        self.assertEqual(len(result), 3)
        self.assertEqual(len(result.columns), 1)

    def test_s3_partial_parquet_file_to_table_when_schema_passed(self):

        pq_file = ParquetFile(PARQUET_FILE_PATH)
        partial_parquet_params = PartialParquetParameters.of(
            pq_metadata=pq_file.metadata
        )
        # only first row group to be downloaded
        partial_parquet_params.row_groups_to_download.pop()

        schema = pa.schema(
            [
                pa.field("n_legs", pa.string()),
                pa.field("animal", pa.string()),
                # NOTE: This field is not in the parquet file, but will be added on as an all-null column
                pa.field("MISSING", pa.int64()),
            ]
        )

        pa_kwargs_provider = lambda content_type, kwargs: {"schema": schema}

        result = s3_partial_parquet_file_to_table(
            PARQUET_FILE_PATH,
            ContentType.PARQUET.value,
            ContentEncoding.IDENTITY.value,
            pa_read_func_kwargs_provider=pa_kwargs_provider,
            partial_file_download_params=partial_parquet_params,
        )

        self.assertEqual(len(result), 3)
        self.assertEqual(len(result.column_names), 3)

        result_schema = result.schema
        self.assertEqual(result_schema.field(0).type, "string")
        self.assertEqual(result_schema.field(0).name, "n_legs")
        self.assertEqual(result_schema.field(1).type, "string")
        self.assertEqual(result_schema.field(1).name, "animal")
        self.assertEqual(result_schema.field(2).type, "int64")
        self.assertEqual(result_schema.field(2).name, "MISSING")

    def test_s3_partial_parquet_file_to_table_when_schema_passed_with_include_columns(
        self,
    ):

        pq_file = ParquetFile(PARQUET_FILE_PATH)
        partial_parquet_params = PartialParquetParameters.of(
            pq_metadata=pq_file.metadata
        )
        # only first row group to be downloaded
        partial_parquet_params.row_groups_to_download.pop()

        schema = pa.schema(
            [pa.field("animal", pa.string()), pa.field("n_legs", pa.string())]
        )

        pa_kwargs_provider = lambda content_type, kwargs: {"schema": schema}

        result = s3_partial_parquet_file_to_table(
            PARQUET_FILE_PATH,
            ContentType.PARQUET.value,
            ContentEncoding.IDENTITY.value,
            ["n_legs", "animal"],
            pa_read_func_kwargs_provider=pa_kwargs_provider,
            partial_file_download_params=partial_parquet_params,
        )

        self.assertEqual(len(result), 3)
        self.assertEqual(len(result.column_names), 2)

        result_schema = result.schema
        self.assertEqual(result_schema.field(0).type, "string")
        self.assertEqual(result_schema.field(0).name, "n_legs")  # order doesn't change

    def test_s3_partial_parquet_file_to_table_when_multiple_row_groups(self):

        pq_file = ParquetFile(PARQUET_FILE_PATH)
        partial_parquet_params = PartialParquetParameters.of(
            pq_metadata=pq_file.metadata
        )

        self.assertEqual(
            partial_parquet_params.num_row_groups, 2, "test_file.parquet has changed."
        )

        result = s3_partial_parquet_file_to_table(
            PARQUET_FILE_PATH,
            content_encoding=ContentEncoding.IDENTITY.value,
            content_type=ContentType.PARQUET.value,
            partial_file_download_params=partial_parquet_params,
        )

        self.assertEqual(len(result), 6)
        self.assertEqual(len(result.columns), 2)


class TestReadCSV(TestCase):
    def test_read_csv_sanity(self):

        schema = pa.schema(
            [("is_active", pa.string()), ("ship_datetime_utc", pa.timestamp("us"))]
        )
        kwargs = content_type_to_reader_kwargs(ContentType.UNESCAPED_TSV.value)
        _add_column_kwargs(
            ContentType.UNESCAPED_TSV.value,
            ["is_active", "ship_datetime_utc"],
            None,
            kwargs,
        )

        read_kwargs_provider = ReadKwargsProviderPyArrowSchemaOverride(schema=schema)

        kwargs = read_kwargs_provider(ContentType.UNESCAPED_TSV.value, kwargs)

        result = pyarrow_read_csv(NON_EMPTY_VALID_UTSV_PATH, **kwargs)

        self.assertEqual(len(result), 3)
        self.assertEqual(len(result.column_names), 2)
        result_schema = result.schema
        for index, field in enumerate(result_schema):
            self.assertEqual(field.name, schema.field(index).name)

        self.assertEqual(result.schema.field(0).type, "string")

    def test_read_csv_when_column_order_changes(self):

        schema = pa.schema(
            [("is_active", pa.string()), ("ship_datetime_utc", pa.timestamp("us"))]
        )
        kwargs = content_type_to_reader_kwargs(ContentType.UNESCAPED_TSV.value)
        _add_column_kwargs(
            ContentType.UNESCAPED_TSV.value,
            ["is_active", "ship_datetime_utc"],
            ["ship_datetime_utc", "is_active"],
            kwargs,
        )

        read_kwargs_provider = ReadKwargsProviderPyArrowSchemaOverride(schema=schema)

        kwargs = read_kwargs_provider(ContentType.UNESCAPED_TSV.value, kwargs)

        result = pyarrow_read_csv(NON_EMPTY_VALID_UTSV_PATH, **kwargs)

        self.assertEqual(len(result), 3)
        self.assertEqual(len(result.column_names), 2)
        result_schema = result.schema
        self.assertEqual(result_schema.field(1).type, "string")
        self.assertEqual(result_schema.field(0).type, "timestamp[us]")

    def test_read_csv_when_partial_columns_included(self):

        schema = pa.schema(
            [("is_active", pa.string()), ("ship_datetime_utc", pa.timestamp("us"))]
        )
        kwargs = content_type_to_reader_kwargs(ContentType.UNESCAPED_TSV.value)
        _add_column_kwargs(
            ContentType.UNESCAPED_TSV.value,
            ["is_active", "ship_datetime_utc"],
            ["is_active"],
            kwargs,
        )

        read_kwargs_provider = ReadKwargsProviderPyArrowSchemaOverride(schema=schema)

        kwargs = read_kwargs_provider(ContentType.UNESCAPED_TSV.value, kwargs)

        result = pyarrow_read_csv(NON_EMPTY_VALID_UTSV_PATH, **kwargs)

        self.assertEqual(len(result), 3)
        self.assertEqual(len(result.column_names), 1)
        result_schema = result.schema
        self.assertEqual(result_schema.field(0).type, "string")

    def test_read_csv_when_column_names_partial(self):

        schema = pa.schema(
            [("is_active", pa.string()), ("ship_datetime_utc", pa.timestamp("us"))]
        )
        kwargs = content_type_to_reader_kwargs(ContentType.UNESCAPED_TSV.value)
        _add_column_kwargs(ContentType.UNESCAPED_TSV.value, ["is_active"], None, kwargs)

        read_kwargs_provider = ReadKwargsProviderPyArrowSchemaOverride(schema=schema)

        kwargs = read_kwargs_provider(ContentType.UNESCAPED_TSV.value, kwargs)

        self.assertRaises(
            pa.lib.ArrowInvalid,
            lambda: pyarrow_read_csv(NON_EMPTY_VALID_UTSV_PATH, **kwargs),
        )

    def test_read_csv_when_empty_csv_sanity(self):

        schema = pa.schema(
            [("is_active", pa.string()), ("ship_datetime_utc", pa.timestamp("us"))]
        )
        kwargs = content_type_to_reader_kwargs(ContentType.UNESCAPED_TSV.value)
        _add_column_kwargs(
            ContentType.UNESCAPED_TSV.value,
            ["is_active", "ship_datetime_utc"],
            None,
            kwargs,
        )

        read_kwargs_provider = ReadKwargsProviderPyArrowSchemaOverride(schema=schema)
        kwargs = read_kwargs_provider(ContentType.UNESCAPED_TSV.value, kwargs)
        result = pyarrow_read_csv(EMPTY_UTSV_PATH, **kwargs)

        self.assertEqual(len(result), 0)
        self.assertEqual(len(result.column_names), 2)
        result_schema = result.schema
        self.assertEqual(result_schema.field(0).type, "string")
        self.assertEqual(result_schema.field(1).type, "timestamp[us]")

    def test_read_csv_when_empty_csv_include_columns(self):

        schema = pa.schema(
            [("is_active", pa.string()), ("ship_datetime_utc", pa.timestamp("us"))]
        )
        kwargs = content_type_to_reader_kwargs(ContentType.UNESCAPED_TSV.value)
        _add_column_kwargs(
            ContentType.UNESCAPED_TSV.value,
            ["is_active", "ship_datetime_utc"],
            ["ship_datetime_utc", "is_active"],
            kwargs,
        )

        read_kwargs_provider = ReadKwargsProviderPyArrowSchemaOverride(schema=schema)

        kwargs = read_kwargs_provider(ContentType.UNESCAPED_TSV.value, kwargs)

        result = pyarrow_read_csv(EMPTY_UTSV_PATH, **kwargs)

        self.assertEqual(len(result), 0)
        self.assertEqual(len(result.column_names), 2)
        result_schema = result.schema
        self.assertEqual(result_schema.field(1).type, "string")
        self.assertEqual(result_schema.field(0).type, "timestamp[us]")

    def test_read_csv_when_empty_csv_include_partial_columns(self):

        schema = pa.schema(
            [("is_active", pa.string()), ("ship_datetime_utc", pa.timestamp("us"))]
        )
        kwargs = content_type_to_reader_kwargs(ContentType.UNESCAPED_TSV.value)
        _add_column_kwargs(
            ContentType.UNESCAPED_TSV.value,
            ["is_active", "ship_datetime_utc"],
            ["is_active"],
            kwargs,
        )

        read_kwargs_provider = ReadKwargsProviderPyArrowSchemaOverride(schema=schema)

        kwargs = read_kwargs_provider(ContentType.UNESCAPED_TSV.value, kwargs)

        result = pyarrow_read_csv(EMPTY_UTSV_PATH, **kwargs)

        self.assertEqual(len(result), 0)
        self.assertEqual(len(result.column_names), 1)
        result_schema = result.schema
        self.assertEqual(result_schema.field(0).type, "string")

    def test_read_csv_when_empty_csv_honors_column_names(self):

        schema = pa.schema(
            [("is_active", pa.string()), ("ship_datetime_utc", pa.timestamp("us"))]
        )
        kwargs = content_type_to_reader_kwargs(ContentType.UNESCAPED_TSV.value)
        _add_column_kwargs(ContentType.UNESCAPED_TSV.value, ["is_active"], None, kwargs)

        read_kwargs_provider = ReadKwargsProviderPyArrowSchemaOverride(schema=schema)

        kwargs = read_kwargs_provider(ContentType.UNESCAPED_TSV.value, kwargs)

        result = pyarrow_read_csv(EMPTY_UTSV_PATH, **kwargs)

        self.assertEqual(len(result), 0)
        self.assertEqual(len(result.column_names), 1)
        result_schema = result.schema
        self.assertEqual(result_schema.field(0).type, "string")

    def test_read_csv_when_empty_csv_and_raise_on_empty_passed(self):

        schema = pa.schema(
            [("is_active", pa.string()), ("ship_datetime_utc", pa.timestamp("us"))]
        )
        kwargs = content_type_to_reader_kwargs(ContentType.UNESCAPED_TSV.value)
        _add_column_kwargs(ContentType.UNESCAPED_TSV.value, ["is_active"], None, kwargs)

        read_kwargs_provider = ReadKwargsProviderPyArrowSchemaOverride(schema=schema)

        kwargs = read_kwargs_provider(ContentType.UNESCAPED_TSV.value, kwargs)

        self.assertRaises(
            pa.lib.ArrowInvalid,
            lambda: pyarrow_read_csv(
                EMPTY_UTSV_PATH, **{**kwargs, RAISE_ON_EMPTY_CSV_KWARG: True}
            ),
        )


class TestS3FileToTable(TestCase):
    def test_s3_file_to_table_identity_sanity(self):

        schema = pa.schema(
            [("is_active", pa.string()), ("ship_datetime_utc", pa.timestamp("us"))]
        )

        result = s3_file_to_table(
            NON_EMPTY_VALID_UTSV_PATH,
            ContentType.UNESCAPED_TSV.value,
            ContentEncoding.IDENTITY.value,
            ["is_active", "ship_datetime_utc"],
            None,
            pa_read_func_kwargs_provider=ReadKwargsProviderPyArrowSchemaOverride(
                schema=schema
            ),
        )

        self.assertEqual(len(result), 3)
        self.assertEqual(len(result.column_names), 2)
        result_schema = result.schema
        for index, field in enumerate(result_schema):
            self.assertEqual(field.name, schema.field(index).name)

        self.assertEqual(result.schema.field(0).type, "string")

    def test_s3_file_to_table_gzip_compressed_sanity(self):

        schema = pa.schema(
            [("is_active", pa.string()), ("ship_datetime_utc", pa.timestamp("us"))]
        )

        result = s3_file_to_table(
            GZIP_COMPRESSED_FILE_UTSV_PATH,
            ContentType.UNESCAPED_TSV.value,
            ContentEncoding.GZIP.value,
            ["is_active", "ship_datetime_utc"],
            None,
            pa_read_func_kwargs_provider=ReadKwargsProviderPyArrowSchemaOverride(
                schema=schema
            ),
        )

        self.assertEqual(len(result), 3)
        self.assertEqual(len(result.column_names), 2)
        result_schema = result.schema
        for index, field in enumerate(result_schema):
            self.assertEqual(field.name, schema.field(index).name)

        self.assertEqual(result.schema.field(0).type, "string")

    def test_s3_file_to_table_bz2_compressed_sanity(self):

        schema = pa.schema(
            [("is_active", pa.string()), ("ship_datetime_utc", pa.timestamp("us"))]
        )

        result = s3_file_to_table(
            BZ2_COMPRESSED_FILE_UTSV_PATH,
            ContentType.UNESCAPED_TSV.value,
            ContentEncoding.BZIP2.value,
            ["is_active", "ship_datetime_utc"],
            None,
            pa_read_func_kwargs_provider=ReadKwargsProviderPyArrowSchemaOverride(
                schema=schema
            ),
        )

        self.assertEqual(len(result), 3)
        self.assertEqual(len(result.column_names), 2)
        result_schema = result.schema
        for index, field in enumerate(result_schema):
            self.assertEqual(field.name, schema.field(index).name)

        self.assertEqual(result.schema.field(0).type, "string")

    def test_s3_file_to_table_when_parquet_sanity(self):

        pa_kwargs_provider = lambda content_type, kwargs: {
            "reader_type": "pyarrow",
            **kwargs,
        }

        result = s3_file_to_table(
            PARQUET_FILE_PATH,
            ContentType.PARQUET.value,
            ContentEncoding.IDENTITY.value,
            ["n_legs", "animal"],
            ["n_legs"],
            pa_read_func_kwargs_provider=pa_kwargs_provider,
        )

        self.assertEqual(len(result), 6)
        self.assertEqual(len(result.column_names), 1)
        schema = result.schema
        schema_index = schema.get_field_index("n_legs")
        self.assertEqual(schema.field(schema_index).type, "int64")

    def test_s3_file_to_table_when_parquet_schema_overridden(self):

        schema = pa.schema(
            [pa.field("animal", pa.string()), pa.field("n_legs", pa.string())]
        )

        pa_kwargs_provider = lambda content_type, kwargs: {
            "schema": schema,
            "reader_type": "pyarrow",
            **kwargs,
        }

        result = s3_file_to_table(
            PARQUET_FILE_PATH,
            ContentType.PARQUET.value,
            ContentEncoding.IDENTITY.value,
            ["n_legs", "animal"],
            pa_read_func_kwargs_provider=pa_kwargs_provider,
        )

        self.assertEqual(len(result), 6)
        self.assertEqual(len(result.column_names), 2)

        result_schema = result.schema
        for index, field in enumerate(result_schema):
            self.assertEqual(field.name, schema.field(index).name)

        self.assertEqual(result.schema.field(1).type, "string")



import logging
from ray import cloudpickle
from collections import defaultdict
import time
from deltacat.io.object_store import IObjectStore
from typing import Any, List, Optional
from deltacat import logs
import uuid
import socket
from pymemcache.client.base import Client
from pymemcache.client.retrying import RetryingClient
from pymemcache.exceptions import MemcacheUnexpectedCloseError
from pymemcache.client.rendezvous import RendezvousHash
from deltacat.utils.cloudpickle import dump_into_chunks

logger = logs.configure_deltacat_logger(logging.getLogger(__name__))


class MemcachedObjectStore(IObjectStore):
    """
    An implementation of object store that uses Memcached.
    """

    CONNECT_TIMEOUT = 10 * 60
    FETCH_TIMEOUT = 30 * 60
    MAX_ITEM_SIZE_BYTES = 1000000000

    def __init__(
        self,
        storage_node_ips: Optional[List[str]] = None,
        port: Optional[int] = 11212,
        connect_timeout: float = CONNECT_TIMEOUT,
        timeout: float = FETCH_TIMEOUT,
        noreply: bool = False,
        max_item_size_bytes: int = MAX_ITEM_SIZE_BYTES,
    ) -> None:
        self.client_cache = {}
        self.current_ip = None
        self.SEPARATOR = "_"
        self.port = port
        self.storage_node_ips = storage_node_ips or []
        for ip in self.storage_node_ips:
            assert (
                self.SEPARATOR not in ip
            ), f"IP address should not contain {self.SEPARATOR}"
        self.hasher = RendezvousHash(nodes=self.storage_node_ips, seed=0)
        self.connect_timeout = connect_timeout
        self.timeout = timeout
        self.noreply = noreply
        self.max_item_size_bytes = max_item_size_bytes
        logger.info(
            f"The storage node IPs: {self.storage_node_ips} with noreply: {self.noreply}"
        )
        super().__init__()

    def put_many(self, objects: List[object], *args, **kwargs) -> List[Any]:
        input = defaultdict(dict)
        result = []
        for obj in objects:
            serialized_list = dump_into_chunks(
                obj, max_size_bytes=self.max_item_size_bytes
            )
            uid = uuid.uuid4()
            create_ref_ip = self._get_create_ref_ip(uid.__str__())

            for chunk_index, chunk in enumerate(serialized_list):
                ref = self._create_ref(uid, create_ref_ip, chunk_index)
                input[create_ref_ip][ref] = chunk

            return_ref = self._create_ref(uid, create_ref_ip, len(serialized_list))
            result.append(return_ref)
        for create_ref_ip, ref_to_object in input.items():
            client = self._get_client_by_ip(create_ref_ip)
            if client.set_many(ref_to_object, noreply=self.noreply):
                raise RuntimeError("Unable to write few keys to cache")

        return result

    def put(self, obj: object, *args, **kwargs) -> Any:
        serialized_list = dump_into_chunks(obj, max_size_bytes=self.max_item_size_bytes)
        uid = uuid.uuid4()
        create_ref_ip = self._get_create_ref_ip(uid.__str__())
        client = self._get_client_by_ip(create_ref_ip)

        for chunk_index, chunk in enumerate(serialized_list):
            ref = self._create_ref(uid, create_ref_ip, chunk_index)

            try:
                if not client.set(ref, chunk, noreply=self.noreply):
                    raise RuntimeError(f"Unable to write {ref} to cache")
            except BaseException as e:
                raise RuntimeError(
                    f"Received {e} while writing ref={ref} and obj size={len(chunk)}"
                )

        return self._create_ref(uid, create_ref_ip, len(serialized_list))

    def get_many(self, refs: List[Any], *args, **kwargs) -> List[object]:
        result = []
        refs_per_ip = defaultdict(lambda: [])
        chunks_by_refs = defaultdict(lambda: [])

        start = time.monotonic()
        for ref in refs:
            uid, ip, chunk_count = ref.split(self.SEPARATOR)
            chunk_count = int(chunk_count)
            for chunk_index in range(chunk_count):
                current_ref = self._create_ref(uid, ip, chunk_index)
                refs_per_ip[ip].append(current_ref)

        total_ref_count = 0
        for (ip, current_refs) in refs_per_ip.items():
            client = self._get_client_by_ip(ip)
            cache_result = client.get_many(current_refs)
            assert len(cache_result) == len(
                current_refs
            ), f"Not all values were returned from cache from {ip} as {len(cache_result)} != {len(current_refs)}"
            total_ref_count += len(cache_result)
            chunks_by_refs.update(cache_result)

        assert len(chunks_by_refs) == total_ref_count, (
            "Total refs retrieved from memcached doesn't "
            f"match as {len(chunks_by_refs)} != {total_ref_count}"
        )

        for ref in refs:
            uid, ip, chunk_count = ref.split(self.SEPARATOR)
            chunk_count = int(chunk_count)

            deserialize_start = time.monotonic()

            chunks = []

            for chunk_index in range(chunk_count):
                current_ref = self._create_ref(uid, ip, chunk_index)
                chunk = chunks_by_refs[current_ref]
                assert (
                    chunk
                ), f"Serialized chunks were not present for ref={current_ref}"
                chunks_by_refs.pop(current_ref)
                chunks.append(chunk)

            assert chunk_count == len(
                chunks
            ), f"The chunk count must be equal for ref={ref}"
            if chunk_count == 1:
                # avoids moving each byte
                serialized = chunks[0]
            else:
                serialized = bytearray()
                for chunk_index in range(chunk_count):
                    chunk = chunks[chunk_index]
                    serialized.extend(chunk)
                    chunks[chunk_index] = None

            deserialized = cloudpickle.loads(serialized)
            result.append(deserialized)

            deserialize_end = time.monotonic()
            logger.debug(
                f"The time taken to deserialize {len(serialized)} bytes"
                f" and {chunk_count} chunks is: {deserialize_end - deserialize_start}",
            )

        end = time.monotonic()

        logger.info(f"The total time taken to read all objects is: {end - start}")

        assert len(result) == len(
            refs
        ), f"The total number of refs must be equal as {len(result)} != {len(refs)}"
        return result

    def get(self, ref: Any, *args, **kwargs) -> object:
        uid, ip, chunk_count = ref.split(self.SEPARATOR)
        chunk_count = int(chunk_count)

        client = self._get_client_by_ip(ip)
        serialized = bytearray()

        for chunk_index in range(chunk_count):
            ref = self._create_ref(uid, ip, chunk_index)
            chunk = client.get(ref)
            serialized.extend(chunk)

        return cloudpickle.loads(serialized)

    def close(self) -> None:
        for client in self.client_cache.values():
            client.close()

        self.client_cache.clear()

    def _create_ref(self, uid, ip, chunk_index) -> str:
        return f"{uid}{self.SEPARATOR}{ip}{self.SEPARATOR}{chunk_index}"

    def _get_storage_node_ip(self, key: str):
        storage_node_ip = self.hasher.get_node(key)
        return storage_node_ip

    def _get_create_ref_ip(self, uid: str):
        if self.storage_node_ips:
            create_ref_ip = self._get_storage_node_ip(uid)
        else:
            create_ref_ip = self._get_current_ip()
        return create_ref_ip

    def _get_client_by_ip(self, ip_address: str):
        if ip_address in self.client_cache:
            return self.client_cache[ip_address]

        base_client = Client(
            (ip_address, self.port),
            connect_timeout=self.connect_timeout,
            timeout=self.timeout,
            no_delay=True,
        )

        client = RetryingClient(
            base_client,
            attempts=15,
            retry_delay=1,
            retry_for=[
                MemcacheUnexpectedCloseError,
                ConnectionRefusedError,
                ConnectionResetError,
                BrokenPipeError,
                TimeoutError,
            ],
        )

        self.client_cache[ip_address] = client
        return client

    def _get_current_ip(self):
        if self.current_ip is None:
            self.current_ip = socket.gethostbyname(socket.gethostname())

        return self.current_ip



from unittest.mock import patch, Mock
import unittest
from http import HTTPStatus
import requests
from requests import HTTPError

HAPPY_RESPONSE = {
    "AccessKeyId": "ASIA123456789",
    "Code": "Success",
    "Expiration": "2023-08-02T13:50:33Z",
    "LastUpdated": "2023-08-02T07:23:35Z",
    "SecretAccessKey": "bar",
    "Token": "foo",
    "Type": "AWS-HMAC",
}


class MockResponse:
    """
    A mock object denoting the response of requests method.
    """

    def __init__(self, status_code: int, text: str, reason: str = "") -> None:
        self.status_code: requests.Response.status_code = status_code
        self.text = text
        self.reason = reason

    def raise_for_status(*args, **kwargs):
        pass


class TestBlockUntilInstanceMetadataServiceReturnsSuccess(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()

    @classmethod
    def tearDownClass(cls) -> None:
        pass

    @patch("deltacat.aws.clients.requests")
    def test_sanity(self, requests_mock):
        from deltacat.aws.clients import (
            block_until_instance_metadata_service_returns_success,
        )

        requests_mock.get.return_value = MockResponse(200, "foo")
        self.assertEqual(
            block_until_instance_metadata_service_returns_success().status_code, 200
        )

    @patch("deltacat.aws.clients.requests")
    def test_retrying_on_statuses_in_status_force_list(self, requests_mock):
        from deltacat.aws.clients import (
            block_until_instance_metadata_service_returns_success,
        )

        requests_mock.get.side_effect = [
            MockResponse(HTTPStatus.OK, "foo"),
            MockResponse(HTTPStatus.TOO_MANY_REQUESTS, "foo"),
            MockResponse(HTTPStatus.INTERNAL_SERVER_ERROR, "foo"),
            MockResponse(HTTPStatus.NOT_IMPLEMENTED, "bar"),
            MockResponse(HTTPStatus.SERVICE_UNAVAILABLE, "bar"),
            MockResponse(HTTPStatus.GATEWAY_TIMEOUT, "bar"),
        ]
        self.assertEqual(
            block_until_instance_metadata_service_returns_success().status_code, 200
        )

    @patch("deltacat.aws.clients.requests.get")
    def test_retrying_on_initial_failures(self, requests_mock_get):
        from deltacat.aws.clients import (
            block_until_instance_metadata_service_returns_success,
        )

        mock_success_response = Mock()
        mock_success_response.raise_for_status.return_value = None
        mock_success_response.status_code = 200

        mock_service_unavailable_response = Mock()
        mock_service_unavailable_response.status_code = 503

        mock_errors = []
        for err in range(3):
            mock_error_response = Mock()
            mock_error_response.raise_for_status.side_effect = HTTPError(
                response=mock_service_unavailable_response
            )
            mock_errors.append(mock_error_response)

        mock_connection_error_response = requests.exceptions.ConnectionError()
        mock_errors.append(mock_connection_error_response)

        requests_mock_get.side_effect = [*mock_errors, mock_success_response]
        self.assertEqual(
            block_until_instance_metadata_service_returns_success().status_code, 200
        )
        self.assertEqual(requests_mock_get.call_count, 5)

    @patch("deltacat.aws.clients.requests")
    def test_retrying_status_on_shortlist_returns_early(self, requests_mock):
        from deltacat.aws.clients import (
            block_until_instance_metadata_service_returns_success,
        )

        requests_mock.get.side_effect = [
            MockResponse(HTTPStatus.FORBIDDEN, "foo"),
        ]
        self.assertEqual(
            block_until_instance_metadata_service_returns_success().status_code, 403
        )



from __future__ import annotations

from ray.types import ObjectRef
from typing import Dict, List, Optional, Any
from deltacat.utils.metrics import MetricsConfig
from deltacat.utils.common import ReadKwargsProvider
from deltacat.io.object_store import IObjectStore
from deltacat.storage import (
    Partition,
    SortKey,
    interface as unimplemented_deltacat_storage,
)
from deltacat.compute.compactor_v2.constants import (
    DROP_DUPLICATES,
    MAX_RECORDS_PER_COMPACTED_FILE,
)
from deltacat.types.media import ContentType
from deltacat.compute.compactor.model.round_completion_info import RoundCompletionInfo
from deltacat.compute.compactor.model.delta_file_envelope import DeltaFileEnvelopeGroups


class MergeInput(Dict):
    @staticmethod
    def of(
        dfe_groups_refs: List[ObjectRef[DeltaFileEnvelopeGroups]],
        write_to_partition: Partition,
        compacted_file_content_type: ContentType,
        primary_keys: List[str],
        hash_group_index: int,
        num_hash_groups: int,
        hash_bucket_count: int,
        drop_duplicates: Optional[bool] = DROP_DUPLICATES,
        sort_keys: Optional[List[SortKey]] = None,
        merge_task_index: Optional[int] = 0,
        max_records_per_output_file: Optional[int] = MAX_RECORDS_PER_COMPACTED_FILE,
        enable_profiler: Optional[bool] = False,
        metrics_config: Optional[MetricsConfig] = None,
        s3_table_writer_kwargs: Optional[Dict[str, Any]] = None,
        read_kwargs_provider: Optional[ReadKwargsProvider] = None,
        round_completion_info: Optional[RoundCompletionInfo] = None,
        object_store: Optional[IObjectStore] = None,
        deltacat_storage=unimplemented_deltacat_storage,
        deltacat_storage_kwargs: Optional[Dict[str, Any]] = None,
    ) -> MergeInput:

        result = MergeInput()
        result["dfe_groups_refs"] = dfe_groups_refs
        result["write_to_partition"] = write_to_partition
        result["compacted_file_content_type"] = compacted_file_content_type
        result["primary_keys"] = primary_keys
        result["hash_group_index"] = hash_group_index
        result["num_hash_groups"] = num_hash_groups
        result["hash_bucket_count"] = hash_bucket_count
        result["drop_duplicates"] = drop_duplicates
        result["sort_keys"] = sort_keys
        result["merge_task_index"] = merge_task_index
        result["max_records_per_output_file"] = max_records_per_output_file
        result["enable_profiler"] = enable_profiler
        result["metrics_config"] = metrics_config
        result["s3_table_writer_kwargs"] = s3_table_writer_kwargs or {}
        result["read_kwargs_provider"] = read_kwargs_provider
        result["round_completion_info"] = round_completion_info
        result["object_store"] = object_store
        result["deltacat_storage"] = deltacat_storage
        result["deltacat_storage_kwargs"] = deltacat_storage_kwargs or {}

        return result

    @property
    def dfe_groups_refs(self) -> List[ObjectRef[DeltaFileEnvelopeGroups]]:
        return self["dfe_groups_refs"]

    @property
    def write_to_partition(self) -> Partition:
        return self["write_to_partition"]

    @property
    def compacted_file_content_type(self) -> ContentType:
        return self["compacted_file_content_type"]

    @property
    def primary_keys(self) -> List[str]:
        return self["primary_keys"]

    @property
    def hash_group_index(self) -> int:
        return self["hash_group_index"]

    @property
    def num_hash_groups(self) -> int:
        return self["num_hash_groups"]

    @property
    def hash_bucket_count(self) -> int:
        return self["hash_bucket_count"]

    @property
    def drop_duplicates(self) -> int:
        return self["drop_duplicates"]

    @property
    def sort_keys(self) -> Optional[List[SortKey]]:
        return self.get("sort_keys")

    @property
    def merge_task_index(self) -> int:
        return self.get("merge_task_index")

    @property
    def max_records_per_output_file(self) -> int:
        return self.get("max_records_per_output_file")

    @property
    def enable_profiler(self) -> bool:
        return self.get("enable_profiler")

    @property
    def metrics_config(self) -> Optional[MetricsConfig]:
        return self.get("metrics_config")

    @property
    def s3_table_writer_kwargs(self) -> Optional[Dict[str, Any]]:
        return self.get("s3_table_writer_kwargs")

    @property
    def read_kwargs_provider(self) -> Optional[ReadKwargsProvider]:
        return self.get("read_kwargs_provider")

    @property
    def round_completion_info(self) -> Optional[RoundCompletionInfo]:
        return self.get("round_completion_info")

    @property
    def object_store(self) -> Optional[IObjectStore]:
        return self.get("object_store")

    @property
    def deltacat_storage(self) -> unimplemented_deltacat_storage:
        return self["deltacat_storage"]

    @property
    def deltacat_storage_kwargs(self) -> Optional[Dict[str, Any]]:
        return self.get("deltacat_storage_kwargs")



from __future__ import annotations

import pyarrow as pa
import pyarrow.fs as pafs
import pyarrow.parquet as papq
import pytest

from deltacat.utils.pyarrow import s3_file_to_table
from deltacat.types.media import (
    ContentEncoding,
    ContentType,
)


def pyarrow_read(path: str, columns: list[str] | None = None) -> pa.Table:
    assert path.startswith(
        "s3://"
    ), f"Expected file path to start with 's3://', but got {path}."
    fs = pafs.S3FileSystem()
    path = path.replace("s3://", "")
    return papq.read_table(path, columns=columns, filesystem=fs)


def deltacat_read(path: str, columns: list[str] | None = None) -> pa.Table:
    assert path.startswith("s3://")
    return s3_file_to_table(
        path,
        content_type=ContentType.PARQUET,
        content_encoding=ContentEncoding.IDENTITY,
        column_names=None,  # Parquet files are schemaful
        include_columns=columns,
    )


def daft_table_read(path: str, columns: list[str] | None = None) -> pa.Table:
    try:
        import daft
    except ImportError:
        raise ImportError(
            "Daft not installed. Install Daft using pip to run these benchmarks: `pip install getdaft`"
        )

    tbl = daft.table.Table.read_parquet(path, columns=columns)
    return tbl.to_arrow()


@pytest.fixture(
    params=[
        daft_table_read,
        pyarrow_read,
        deltacat_read,
    ],
    ids=[
        "daft_table",
        "pyarrow",
        "deltacat",
    ],
)
def read_fn(request):
    """Fixture which returns the function to read a PyArrow table from a path"""
    return request.param



import logging
import importlib
from deltacat.compute.compactor_v2.model.merge_input import MergeInput
import numpy as np
import pyarrow as pa
import ray
import time
import pyarrow.compute as pc
from uuid import uuid4
from collections import defaultdict
from deltacat import logs
from typing import List, Optional
from deltacat.types.media import DELIMITED_TEXT_CONTENT_TYPES
from deltacat.compute.compactor_v2.model.merge_result import MergeResult
from deltacat.compute.compactor.model.materialize_result import MaterializeResult
from deltacat.compute.compactor.model.pyarrow_write_result import PyArrowWriteResult
from deltacat.compute.compactor import (
    RoundCompletionInfo,
    DeltaFileEnvelope,
)
from deltacat.utils.common import ReadKwargsProvider

from contextlib import nullcontext
from deltacat.types.tables import TABLE_CLASS_TO_SIZE_FUNC
from deltacat.utils.ray_utils.runtime import (
    get_current_ray_task_id,
    get_current_ray_worker_id,
)
from deltacat.compute.compactor.utils import system_columns as sc

from deltacat.utils.performance import timed_invocation
from deltacat.utils.metrics import emit_timer_metrics
from deltacat.utils.resources import get_current_node_peak_memory_usage_in_bytes
from deltacat.compute.compactor_v2.utils.primary_key_index import (
    generate_pk_hash_column,
    hash_group_index_to_hash_bucket_indices,
)
from deltacat.storage import (
    Delta,
    DeltaLocator,
    DeltaType,
    Manifest,
    Partition,
    interface as unimplemented_deltacat_storage,
)
from deltacat.compute.compactor_v2.utils.dedupe import drop_duplicates


if importlib.util.find_spec("memray"):
    import memray

logger = logs.configure_deltacat_logger(logging.getLogger(__name__))


def _append_delta_type_column(table: pa.Table, value: np.bool_):
    return table.append_column(
        sc._DELTA_TYPE_COLUMN_FIELD,
        pa.array(np.repeat(value, len(table)), sc._DELTA_TYPE_COLUMN_TYPE),
    )


def _drop_delta_type_rows(table: pa.Table, delta_type: DeltaType) -> pa.Table:
    if sc._DELTA_TYPE_COLUMN_NAME not in table.column_names:
        return table

    delta_type_value = sc.delta_type_to_field(delta_type)

    result = table.filter(
        pc.not_equal(table[sc._DELTA_TYPE_COLUMN_NAME], delta_type_value)
    )

    return result.drop([sc._DELTA_TYPE_COLUMN_NAME])


def _build_incremental_table(
    hash_bucket_index: int,
    df_envelopes_list: List[List[DeltaFileEnvelope]],
) -> pa.Table:

    logger.info(
        f"[Hash bucket index {hash_bucket_index}] Reading dedupe input for "
        f"{len(df_envelopes_list)} delta file envelope lists..."
    )
    hb_tables = []
    # sort by delta file stream position now instead of sorting every row later
    df_envelopes = [d for dfe_list in df_envelopes_list for d in dfe_list]
    df_envelopes = sorted(
        df_envelopes,
        key=lambda df: (df.stream_position, df.file_index),
        reverse=False,  # ascending
    )
    is_delete = False
    for df_envelope in df_envelopes:
        assert (
            df_envelope.delta_type != DeltaType.APPEND
        ), "APPEND type deltas are not supported. Kindly use UPSERT or DELETE"
        if df_envelope.delta_type == DeltaType.DELETE:
            is_delete = True

    for df_envelope in df_envelopes:
        table = df_envelope.table
        if is_delete:
            table = _append_delta_type_column(
                table, np.bool_(sc.delta_type_to_field(df_envelope.delta_type))
            )

        hb_tables.append(table)

    result = pa.concat_tables(hb_tables)

    return result


def _merge_tables(
    table: pa.Table,
    primary_keys: List[str],
    can_drop_duplicates: bool,
    compacted_table: Optional[pa.Table] = None,
) -> pa.Table:
    """
    Merges the table with compacted table dropping duplicates where necessary.

    This method ensures the appropriate deltas of types DELETE/UPSERT are correctly
    appended to the table.
    """

    all_tables = []
    incremental_idx = 0

    if compacted_table:
        incremental_idx = 1
        all_tables.append(compacted_table)

    all_tables.append(table)

    if not primary_keys or not can_drop_duplicates:
        logger.info(
            f"Not dropping duplicates for primary keys={primary_keys} "
            f"and can_drop_duplicates={can_drop_duplicates}"
        )
        all_tables[incremental_idx] = _drop_delta_type_rows(
            all_tables[incremental_idx], DeltaType.DELETE
        )
        # we need not drop duplicates
        return pa.concat_tables(all_tables)

    all_tables = generate_pk_hash_column(all_tables, primary_keys=primary_keys)

    result_table_list = []

    incremental_table = drop_duplicates(
        all_tables[incremental_idx], on=sc._PK_HASH_STRING_COLUMN_NAME
    )

    if compacted_table:
        compacted_table = all_tables[0]

        records_to_keep = pc.invert(
            pc.is_in(
                compacted_table[sc._PK_HASH_STRING_COLUMN_NAME],
                incremental_table[sc._PK_HASH_STRING_COLUMN_NAME],
            )
        )

        result_table_list.append(compacted_table.filter(records_to_keep))

    incremental_table = _drop_delta_type_rows(incremental_table, DeltaType.DELETE)
    result_table_list.append(incremental_table)

    final_table = pa.concat_tables(result_table_list)
    final_table = final_table.drop([sc._PK_HASH_STRING_COLUMN_NAME])

    return final_table


def _download_compacted_table(
    hb_index: int,
    rcf: RoundCompletionInfo,
    read_kwargs_provider: Optional[ReadKwargsProvider] = None,
    deltacat_storage=unimplemented_deltacat_storage,
    deltacat_storage_kwargs: Optional[dict] = None,
) -> pa.Table:
    tables = []
    hb_index_to_indices = rcf.hb_index_to_entry_range

    if str(hb_index) not in hb_index_to_indices:
        return None

    indices = hb_index_to_indices[str(hb_index)]

    assert (
        indices is not None and len(indices) == 2
    ), "indices should not be none and contains exactly two elements"

    for offset in range(indices[1] - indices[0]):
        table = deltacat_storage.download_delta_manifest_entry(
            rcf.compacted_delta_locator,
            entry_index=(indices[0] + offset),
            file_reader_kwargs_provider=read_kwargs_provider,
            **deltacat_storage_kwargs,
        )

        tables.append(table)

    return pa.concat_tables(tables)


def _copy_all_manifest_files_from_old_hash_buckets(
    hb_index_copy_by_reference: List[int],
    round_completion_info: RoundCompletionInfo,
    write_to_partition: Partition,
    deltacat_storage=unimplemented_deltacat_storage,
    deltacat_storage_kwargs: Optional[dict] = None,
) -> List[MaterializeResult]:

    compacted_delta_locator = round_completion_info.compacted_delta_locator
    manifest = deltacat_storage.get_delta_manifest(
        compacted_delta_locator, **deltacat_storage_kwargs
    )

    manifest_entry_referenced_list = []
    materialize_result_list = []
    hb_index_to_indices = round_completion_info.hb_index_to_entry_range

    if hb_index_to_indices is None:
        logger.info(f"Nothing to copy by reference. Skipping...")
        return []

    for hb_index in hb_index_copy_by_reference:
        if str(hb_index) not in hb_index_to_indices:
            continue

        indices = hb_index_to_indices[str(hb_index)]
        for offset in range(indices[1] - indices[0]):
            entry_index = indices[0] + offset
            assert entry_index < len(
                manifest.entries
            ), f"entry index: {entry_index} >= {len(manifest.entries)}"
            manifest_entry = manifest.entries[entry_index]
            manifest_entry_referenced_list.append(manifest_entry)

        manifest = Manifest.of(
            entries=manifest_entry_referenced_list, uuid=str(uuid4())
        )
        delta = Delta.of(
            locator=DeltaLocator.of(write_to_partition.locator),
            delta_type=DeltaType.UPSERT,
            meta=manifest.meta,
            manifest=manifest,
            previous_stream_position=write_to_partition.stream_position,
            properties={},
        )
        referenced_pyarrow_write_result = PyArrowWriteResult.of(
            len(manifest_entry_referenced_list),
            manifest.meta.source_content_length,
            manifest.meta.content_length,
            manifest.meta.record_count,
        )
        materialize_result = MaterializeResult.of(
            delta=delta,
            task_index=hb_index,
            pyarrow_write_result=referenced_pyarrow_write_result,
            referenced_pyarrow_write_result=referenced_pyarrow_write_result,
        )
        materialize_result_list.append(materialize_result)
    return materialize_result_list


def _timed_merge(input: MergeInput) -> MergeResult:
    def _materialize(
        hash_bucket_index,
        compacted_tables: List[pa.Table],
    ) -> MaterializeResult:
        compacted_table = pa.concat_tables(compacted_tables)
        if input.compacted_file_content_type in DELIMITED_TEXT_CONTENT_TYPES:
            # TODO (rkenmi): Investigate if we still need to convert this table to pandas DataFrame
            # TODO (pdames): compare performance to pandas-native materialize path
            df = compacted_table.to_pandas(split_blocks=True, self_destruct=True)
            compacted_table = df
        delta, stage_delta_time = timed_invocation(
            input.deltacat_storage.stage_delta,
            compacted_table,
            input.write_to_partition,
            max_records_per_entry=input.max_records_per_output_file,
            content_type=input.compacted_file_content_type,
            s3_table_writer_kwargs=input.s3_table_writer_kwargs,
            **input.deltacat_storage_kwargs,
        )
        compacted_table_size = TABLE_CLASS_TO_SIZE_FUNC[type(compacted_table)](
            compacted_table
        )
        logger.debug(
            f"Time taken for materialize task"
            f" to upload {len(compacted_table)} records"
            f" of size {compacted_table_size} is: {stage_delta_time}s"
        )
        manifest = delta.manifest
        manifest_records = manifest.meta.record_count
        assert manifest_records == len(compacted_table), (
            f"Unexpected Error: Materialized delta manifest record count "
            f"({manifest_records}) does not equal compacted table record count "
            f"({len(compacted_table)})"
        )
        materialize_result = MaterializeResult.of(
            delta=delta,
            task_index=hash_bucket_index,
            # TODO (pdames): Generalize WriteResult to contain in-memory-table-type
            #  and in-memory-table-bytes instead of tight coupling to paBytes
            pyarrow_write_result=PyArrowWriteResult.of(
                len(manifest.entries),
                TABLE_CLASS_TO_SIZE_FUNC[type(compacted_table)](compacted_table),
                manifest.meta.content_length,
                len(compacted_table),
            ),
        )
        logger.info(f"Materialize result: {materialize_result}")
        return materialize_result

    task_id = get_current_ray_task_id()
    worker_id = get_current_ray_worker_id()
    with memray.Tracker(
        f"merge_{worker_id}_{task_id}.bin"
    ) if input.enable_profiler else nullcontext():
        # In V2, we need to mitigate risk of running out of memory here in cases of
        #  severe skew of primary key updates in deltas. By severe skew, we mean
        #  one hash bucket require more memory than a worker instance have.
        logger.info(
            f"[Merge task {input.merge_task_index}] Getting delta file envelope "
            f"groups for {len(input.dfe_groups_refs)} object refs..."
        )

        delta_file_envelope_groups_list = input.object_store.get_many(
            input.dfe_groups_refs
        )
        hb_index_to_delta_file_envelopes_list = defaultdict(list)
        for delta_file_envelope_groups in delta_file_envelope_groups_list:
            assert input.hash_bucket_count == len(delta_file_envelope_groups), (
                f"The hash bucket count must match the dfe size as {input.hash_bucket_count}"
                f" != {len(delta_file_envelope_groups)}"
            )

            for hb_idx, dfes in enumerate(delta_file_envelope_groups):
                if dfes:
                    hb_index_to_delta_file_envelopes_list[hb_idx].append(dfes)

        valid_hb_indices_iterable = hash_group_index_to_hash_bucket_indices(
            input.hash_group_index, input.hash_bucket_count, input.num_hash_groups
        )

        total_deduped_records = 0
        total_dfes_found = 0

        materialized_results: List[MaterializeResult] = []
        hb_index_copy_by_reference = []
        for hb_idx in valid_hb_indices_iterable:
            dfe_list = hb_index_to_delta_file_envelopes_list.get(hb_idx)

            if dfe_list:
                total_dfes_found += 1
                table = _build_incremental_table(hb_idx, dfe_list)

                incremental_len = len(table)
                logger.info(
                    f"Got the incremental table of length {incremental_len} for hash bucket {hb_idx}"
                )

                if input.sort_keys:
                    # Incremental is sorted and merged, as sorting
                    # on non event based sort key does not produce consistent
                    # compaction results. E.g., compaction(delta1, delta2, delta3)
                    # will not be equal to compaction(compaction(delta1, delta2), delta3).
                    table = table.sort_by(input.sort_keys)

                compacted_table = None
                if (
                    input.round_completion_info
                    and input.round_completion_info.hb_index_to_entry_range
                    and input.round_completion_info.hb_index_to_entry_range.get(
                        str(hb_idx)
                    )
                    is not None
                ):

                    compacted_table = _download_compacted_table(
                        hb_index=hb_idx,
                        rcf=input.round_completion_info,
                        read_kwargs_provider=input.read_kwargs_provider,
                        deltacat_storage=input.deltacat_storage,
                        deltacat_storage_kwargs=input.deltacat_storage_kwargs,
                    )

                hb_table_record_count = len(table) + (
                    len(compacted_table) if compacted_table else 0
                )

                table, merge_time = timed_invocation(
                    func=_merge_tables,
                    table=table,
                    primary_keys=input.primary_keys,
                    can_drop_duplicates=input.drop_duplicates,
                    compacted_table=compacted_table,
                )
                total_deduped_records += hb_table_record_count - len(table)

                logger.info(
                    f"[Merge task index {input.merge_task_index}] Merged "
                    f"record count: {len(table)}, size={table.nbytes} took: {merge_time}s"
                )

                materialized_results.append(_materialize(hb_idx, [table]))
            else:
                hb_index_copy_by_reference.append(hb_idx)

        if input.round_completion_info and hb_index_copy_by_reference:
            referenced_materialized_results = (
                _copy_all_manifest_files_from_old_hash_buckets(
                    hb_index_copy_by_reference,
                    input.round_completion_info,
                    input.write_to_partition,
                    input.deltacat_storage,
                    input.deltacat_storage_kwargs,
                )
            )
            logger.info(
                f"Copying {len(referenced_materialized_results)} manifest files by reference..."
            )
            materialized_results.extend(referenced_materialized_results)

        logger.info(
            "Total number of materialized results produced for "
            f"hash group index: {input.hash_group_index} is {len(materialized_results)}"
        )

        assert total_dfes_found == len(hb_index_to_delta_file_envelopes_list), (
            "The total dfe list does not match the input dfes from hash bucket as "
            f"{total_dfes_found} != {len(hb_index_to_delta_file_envelopes_list)}"
        )

        peak_memory_usage_bytes = get_current_node_peak_memory_usage_in_bytes()

        return MergeResult(
            materialized_results,
            np.int64(total_deduped_records),
            np.double(peak_memory_usage_bytes),
            np.double(0.0),
            np.double(time.time()),
        )


@ray.remote
def merge(input: MergeInput) -> MergeResult:

    logger.info(f"Starting merge task...")
    merge_result, duration = timed_invocation(func=_timed_merge, input=input)

    emit_metrics_time = 0.0
    if input.metrics_config:
        emit_result, latency = timed_invocation(
            func=emit_timer_metrics,
            metrics_name="merge",
            value=duration,
            metrics_config=input.metrics_config,
        )
        emit_metrics_time = latency

    logger.info(f"Finished merge task...")
    return MergeResult(
        merge_result[0],
        merge_result[1],
        merge_result[2],
        np.double(emit_metrics_time),
        merge_result[4],
    )



from deltacat.utils.common import env_string

# Environment variables
DELTACAT_SYS_LOG_LEVEL = env_string("DELTACAT_SYS_LOG_LEVEL", "DEBUG")
DELTACAT_SYS_LOG_DIR = env_string(
    "DELTACAT_SYS_LOG_DIR",
    "/tmp/deltacat/var/output/logs/",
)
DELTACAT_SYS_INFO_LOG_BASE_FILE_NAME = env_string(
    "DELTACAT_SYS_INFO_LOG_BASE_FILE_NAME",
    "deltacat-python.info.log",
)
DELTACAT_SYS_DEBUG_LOG_BASE_FILE_NAME = env_string(
    "DELTACAT_SYS_DEBUG_LOG_BASE_FILE_NAME",
    "deltacat-python.debug.log",
)

DELTACAT_APP_LOG_LEVEL = env_string("DELTACAT_APP_LOG_LEVEL", "DEBUG")
DELTACAT_APP_LOG_DIR = env_string(
    "DELTACAT_APP_LOG_DIR",
    "/tmp/deltacat/var/output/logs/",
)
DELTACAT_APP_INFO_LOG_BASE_FILE_NAME = env_string(
    "DELTACAT_APP_INFO_LOG_BASE_FILE_NAME",
    "application.info.log",
)
DELTACAT_APP_DEBUG_LOG_BASE_FILE_NAME = env_string(
    "DELTACAT_APP_DEBUG_LOG_BASE_FILE_NAME",
    "application.debug.log",
)

# Byte Units
BYTES_PER_KIBIBYTE = 2**10
BYTES_PER_MEBIBYTE = 2**20
BYTES_PER_GIBIBYTE = 2**30
BYTES_PER_TEBIBYTE = 2**40
BYTES_PER_PEBIBYTE = 2**50

SIGNED_INT64_MIN_VALUE = -(2**63)
SIGNED_INT64_MAX_VALUE = 2**63 - 1

# Inflation multiplier from snappy-compressed parquet to pyarrow.
# This should be kept larger than actual average inflation multipliers.
# Note that this is a very rough guess since actual observed pyarrow
# inflation multiplier for snappy-compressed parquet is about 5.45X for
# all rows, but here we're trying to guess the inflation multipler for just
# a primary key SHA1 digest and sort key columns (which could be all columns
# of the table in the worst case, but here we're assuming that they
# represent no more than ~1/4th of the total table bytes)
PYARROW_INFLATION_MULTIPLIER = 2.5

# Inflation multiplier from snappy-compressed parquet to pyarrow for all columns.
PYARROW_INFLATION_MULTIPLIER_ALL_COLUMNS = 6

MEMORY_TO_HASH_BUCKET_COUNT_RATIO = 0.0512 * BYTES_PER_TEBIBYTE



import unittest
from unittest import mock


class TestRedisObjectStore(unittest.TestCase):

    TEST_VALUE = "test-value"

    def setUp(self):
        self.cloudpickle_patcher = mock.patch(
            "deltacat.io.redis_object_store.cloudpickle"
        )
        self.cloudpickle_mock = self.cloudpickle_patcher.start()
        self.socket_patcher = mock.patch("deltacat.io.redis_object_store.socket")
        self.socket_mock = self.socket_patcher.start()

        self.cloudpickle_mock.dumps.return_value = self.TEST_VALUE
        self.cloudpickle_mock.loads.return_value = self.TEST_VALUE
        self.socket_mock.gethostbyname.return_value = "0.0.0.0"
        self.socket_mock.gethostname.return_value = "test-host"

        from deltacat.io.redis_object_store import RedisObjectStore

        self.object_store = RedisObjectStore()

        super().setUpClass()

    def tearDown(self) -> None:
        self.cloudpickle_patcher.stop()
        self.socket_patcher.stop()

    @mock.patch("deltacat.io.redis_object_store.redis")
    def test_put_many_sanity(self, mock_client):
        mock_client.Redis.return_value.mset.return_value = ["a", "b"]

        result = self.object_store.put_many(["a", "b"])

        self.assertEqual(2, len(result))
        self.assertRegex(result[0], ".*_.*")
        self.assertEqual(1, mock_client.Redis.return_value.mset.call_count)

    @mock.patch("deltacat.io.redis_object_store.redis")
    def test_put_many_when_cache_fails(self, mock_client):
        mock_client.Redis.return_value.mset.return_value = []

        with self.assertRaises(RuntimeError):
            self.object_store.put_many(["a", "b"])

        self.assertEqual(1, mock_client.Redis.return_value.mset.call_count)

    @mock.patch("deltacat.io.redis_object_store.redis")
    def test_get_many_sanity(self, mock_client):
        mock_client.Redis.return_value.mget.return_value = ["a", "b"]

        result = self.object_store.get_many(["test_ip", "test_ip"])

        self.assertEqual(2, len(result))
        self.assertEqual(1, mock_client.Redis.return_value.mget.call_count)

    @mock.patch("deltacat.io.redis_object_store.redis")
    def test_get_many_when_cache_expired(self, mock_client):
        mock_client.Redis.return_value.mget.return_value = ["value1"]

        with self.assertRaises(AssertionError):
            self.object_store.get_many(["test_ip", "test_ip"])

        self.assertEqual(1, mock_client.Redis.return_value.mget.call_count)

    @mock.patch("deltacat.io.redis_object_store.redis")
    def test_get_sanity(self, mock_client):
        mock_client.Redis.return_value.get.return_value = self.TEST_VALUE

        result = self.object_store.get("test_ip")

        self.assertEqual(self.TEST_VALUE, result)
        self.assertEqual(1, mock_client.Redis.return_value.get.call_count)

    @mock.patch("deltacat.io.redis_object_store.redis")
    def test_get_when_cache_fails(self, mock_client):
        mock_client.Redis.return_value.get.side_effect = RuntimeError()

        with self.assertRaises(RuntimeError):
            self.object_store.get("test_ip")

        self.assertEqual(1, mock_client.Redis.return_value.get.call_count)

    @mock.patch("deltacat.io.redis_object_store.redis")
    def test_put_sanity(self, mock_client):
        mock_client.Redis.return_value.set.return_value = True

        result = self.object_store.put("test")

        self.assertIsNotNone(result)
        self.assertEqual(1, mock_client.Redis.return_value.set.call_count)

    @mock.patch("deltacat.io.redis_object_store.redis")
    def test_put_when_cache_fails(self, mock_client):
        mock_client.Redis.return_value.set.return_value = False

        with self.assertRaises(RuntimeError):
            self.object_store.put("test_ip")

        self.assertEqual(1, mock_client.Redis.return_value.set.call_count)



import unittest
from unittest import mock
import time


class TestGetCurrentClusterUtilization(unittest.TestCase):
    @mock.patch("deltacat.utils.resources.ray")
    def test_sanity(self, ray_mock):
        ray_mock.cluster_resources.return_value = {
            "CPU": 10,
            "memory": 10,
            "object_store_memory": 5,
        }
        ray_mock.available_resources.return_value = {
            "CPU": 6,
            "memory": 4,
            "object_store_memory": 5,
        }

        from deltacat.utils.resources import ClusterUtilization

        result = ClusterUtilization.get_current_cluster_utilization()

        self.assertEqual(10, result.total_cpu)
        self.assertEqual(4, result.used_cpu)
        self.assertEqual(10, result.total_memory_bytes)
        self.assertEqual(5, result.total_object_store_memory_bytes)
        self.assertEqual(0, result.used_object_store_memory_bytes)
        self.assertEqual(6, result.used_memory_bytes)
        self.assertIsNotNone(result.used_resources)


class TestClusterUtilizationOverTimeRange(unittest.TestCase):
    @mock.patch("deltacat.utils.resources.ray")
    def test_sanity(self, ray_mock):
        from deltacat.utils.resources import ClusterUtilizationOverTimeRange

        ray_mock.cluster_resources.side_effect = [{"CPU": 32} for _ in range(5)]
        ray_mock.available_resources.side_effect = [
            {"CPU": 2 ** (i + 1)} for i in range(5)
        ]

        with ClusterUtilizationOverTimeRange() as cu:
            time.sleep(3)
            self.assertTrue(cu.used_vcpu_seconds <= 82)  # 30 + 28 + 24
            self.assertTrue(
                cu.total_vcpu_seconds >= cu.used_vcpu_seconds
            )  # total is greater than used
            self.assertIsNotNone(cu.total_memory_gb_seconds)
            self.assertIsNotNone(cu.used_memory_gb_seconds)
            self.assertIsNotNone(cu.max_cpu)



import logging
import os
import re

import setuptools

logger = logging.getLogger(__name__)

ROOT_DIR = os.path.dirname(__file__)


def find_version(*paths):
    version_file_path = os.path.join(ROOT_DIR, *paths)
    with open(version_file_path) as file_stream:
        version_match = re.search(
            r"^__version__ = ['\"]([^'\"]*)['\"]", file_stream.read(), re.M
        )
        if version_match:
            return version_match.group(1)
        raise RuntimeError(f"Failed to find version at: {version_file_path}")


with open(os.path.join(ROOT_DIR, "README.md"), "r", encoding="utf-8") as fh:
    long_description = fh.read()


setuptools.setup(
    name="deltacat",
    version=find_version("deltacat", "__init__.py"),
    author="Ray Team",
    description="A scalable, fast, ACID-compliant Data Catalog powered by Ray.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/ray-project/deltacat",
    packages=setuptools.find_packages(where=".", include="deltacat*"),
    install_requires=[
        # any changes here should also be reflected in requirements.txt
        "aws-embedded-metrics == 3.2.0",
        "boto3 ~= 1.20",
        "numpy == 1.21.5",
        "pandas == 1.3.5",
        "pyarrow == 12.0.1",
        "pydantic == 1.10.4",
        "ray[default] ~= 2.0",
        "s3fs == 2022.2.0",
        "tenacity == 8.1.0",
        "typing-extensions == 4.4.0",
        "pymemcache == 4.0.0",
        "redis == 4.6.0",
        "getdaft == 0.1.17",
        "schedule == 1.2.0",
    ],
    setup_requires=["wheel"],
    package_data={
        "compute/metastats": ["*.yaml"],
    },
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.7",
)



import pyarrow as pa


def coerce_pyarrow_table_to_schema(
    pa_table: pa.Table, input_schema: pa.Schema
) -> pa.Table:
    """Coerces a PyArrow table to the supplied schema

    1. For each field in `pa_table`, cast it to the field in `input_schema` if one with a matching name
        is available
    2. Reorder the fields in the casted table to the supplied schema, dropping any fields in `pa_table`
        that do not exist in the supplied schema
    3. If any fields in the supplied schema are not present, add a null array of the correct type

    Args:
        pa_table (pa.Table): Table to coerce
        input_schema (pa.Schema): Schema to coerce to

    Returns:
        pa.Table: Table with schema == `input_schema`
    """
    input_schema_names = set(input_schema.names)

    # Perform casting of types to provided schema's types
    cast_to_schema = [
        input_schema.field(inferred_field.name)
        if inferred_field.name in input_schema_names
        else inferred_field
        for inferred_field in pa_table.schema
    ]
    casted_table = pa_table.cast(pa.schema(cast_to_schema))

    # Reorder and pad columns with a null column where necessary
    pa_table_column_names = set(casted_table.column_names)
    columns = []
    for name in input_schema.names:
        if name in pa_table_column_names:
            columns.append(casted_table[name])
        else:
            columns.append(
                pa.nulls(len(casted_table), type=input_schema.field(name).type)
            )
    return pa.table(columns, schema=input_schema)



import time
from typing import Any, Callable, Tuple

from deltacat.utils.ray_utils.collections import DistributedCounter


def invoke_with_perf_counter(
    counter: DistributedCounter, counter_key: Any, func: Callable, *args, **kwargs
) -> Tuple[Any, float]:

    start = time.perf_counter()
    result = func(*args, **kwargs)
    stop = time.perf_counter()
    latency = stop - start
    counter.increment.remote(counter_key, latency)
    return result, latency



import logging
from collections import defaultdict
from typing import Any, Dict, List, Optional

import pyarrow
import ray

from deltacat import LocalTable, TableType, logs
from deltacat.compute.stats.models.delta_column_stats import DeltaColumnStats
from deltacat.compute.stats.models.delta_stats import DeltaStats, DeltaStatsCacheMiss
from deltacat.compute.stats.models.delta_stats_cache_result import DeltaStatsCacheResult
from deltacat.compute.stats.models.manifest_entry_stats import ManifestEntryStats
from deltacat.compute.stats.models.stats_result import StatsResult
from deltacat.compute.stats.utils.intervals import DeltaRange
from deltacat.compute.stats.utils.manifest_stats_file import (
    read_manifest_stats_by_columns,
    write_manifest_stats_file,
)
from deltacat.storage import Delta, DeltaLocator, PartitionLocator
from deltacat.storage import interface as unimplemented_deltacat_storage

logger = logs.configure_deltacat_logger(logging.getLogger(__name__))


@ray.remote
def read_cached_delta_stats(
    delta: Delta, columns_to_fetch: List[str], stat_results_s3_bucket: str
):
    """Read delta stats that are cached in S3

    This Ray distributed task reads delta stats from a file system (i.e. S3) based on specified columns.
    Stats are extracted from columns that are found (cache hit), while columns that are missing
    from the file system will record their column names and the delta locator as a cache miss.

    Args:
        delta: The delta object to look up
        columns_to_fetch: Columns to look up for this delta
        stat_results_s3_bucket: The S3 bucket name
    """

    delta_locator = DeltaLocator.of(delta.partition_locator, delta.stream_position)
    column_stats_completion_info: List[
        DeltaColumnStats
    ] = read_manifest_stats_by_columns(
        stat_results_s3_bucket, columns_to_fetch, delta_locator
    )

    found_columns_stats: List[DeltaColumnStats] = []
    missed_columns: List[str] = []
    for column_stats in column_stats_completion_info:
        if column_stats.manifest_stats:
            found_columns_stats.append(column_stats)
        else:
            missed_columns.append(column_stats.column)

    found_stats: Optional[DeltaStats] = (
        DeltaStats.of(found_columns_stats) if found_columns_stats else None
    )
    missed_stats: Optional[DeltaStatsCacheMiss] = (
        DeltaStatsCacheMiss(missed_columns, delta.locator) if missed_columns else None
    )

    return DeltaStatsCacheResult.of(found_stats, missed_stats)


@ray.remote
def cache_delta_column_stats(
    stat_results_s3_bucket: str, dataset_column: DeltaColumnStats
) -> None:
    """Ray distributed task to cache the delta column stats into a file system (i.e. S3).

    Args:
        stat_results_s3_bucket: The S3 bucket name
        dataset_column: Column-oriented stats for a given delta
    """
    write_manifest_stats_file(
        stat_results_s3_bucket, dataset_column.column, dataset_column.manifest_stats
    )


@ray.remote
def get_delta_stats(
    delta_locator: DeltaLocator,
    columns: Optional[List[str]] = None,
    deltacat_storage=unimplemented_deltacat_storage,
    deltacat_storage_kwargs: Optional[Dict[str, Any]] = None,
) -> DeltaStats:
    """Ray distributed task to compute and collect stats for a requested delta.
    If no columns are requested, stats will be computed for all columns.
    Args:
        delta_locator: A reference to the delta
        columns: Column names to specify for this delta. If not provided, all columns are considered.
        deltacat_storage: Client implementation of the DeltaCAT storage interface
    Returns:
        A delta wide stats container
    """
    if deltacat_storage_kwargs is None:
        deltacat_storage_kwargs = {}
    manifest = deltacat_storage.get_delta_manifest(
        delta_locator, **deltacat_storage_kwargs
    )
    delta = Delta.of(delta_locator, None, None, None, manifest)
    return _collect_stats_by_columns(
        delta, columns, deltacat_storage, deltacat_storage_kwargs
    )


@ray.remote
def get_deltas_from_range(
    source_partition_locator: PartitionLocator,
    start_position_inclusive: DeltaRange,
    end_position_inclusive: DeltaRange,
    deltacat_storage=unimplemented_deltacat_storage,
    **kwargs,
) -> List[Delta]:
    """Looks up deltas in the specified partition using Ray, given both starting and ending delta stream positions.

    Args:
        source_partition_locator: Reference to the partition locator tied to the given delta stream positions
        start_position_inclusive: Starting stream position of a range interval.
            Can be an int type to represent a closed bounded range, or a None type to represent unbounded infinity.
        end_position_inclusive: Ending stream position of a range interval.
            Can be an int type to represent a closed bounded range, or a None type to represent unbounded infinity.
        deltacat_storage: Client implementation of the DeltaCAT storage interface

    Returns:
        a list of delta objects
    """

    namespace, partition_values = (
        source_partition_locator.namespace,
        source_partition_locator.partition_values,
    )
    table_name, table_version = (
        source_partition_locator.table_name,
        source_partition_locator.table_version,
    )
    deltas_list_result = deltacat_storage.list_deltas(
        namespace,
        table_name,
        partition_values,
        table_version,
        start_position_inclusive,
        end_position_inclusive,
        ascending_order=True,
        include_manifest=False,
        **kwargs,
    )
    return deltas_list_result.all_items()


def _collect_stats_by_columns(
    delta: Delta,
    columns_to_compute: Optional[List[str]] = None,
    deltacat_storage=unimplemented_deltacat_storage,
    deltacat_storage_kwargs: Optional[Dict[str, Any]] = None,
) -> DeltaStats:
    """Materializes one manifest entry at a time to save memory usage and calculate stats from each of its columns.
    Args:
        delta: A delta object to calculate stats for
        columns_to_compute: Columns to calculate stats for. If not provided, all columns are considered.
        deltacat_storage: Client implementation of the DeltaCAT storage interface
    Returns:
        A delta wide stats container
    """
    if deltacat_storage_kwargs is None:
        deltacat_storage_kwargs = {}
    assert (
        delta.manifest is not None
    ), f"Manifest should not be missing from delta for stats calculation: {delta}"

    # Mapping of column_name -> [stats_file_idx_1, stats_file_idx_2, ... stats_file_idx_n]
    column_stats_map: Dict[str, List[Optional[StatsResult]]] = defaultdict(
        lambda: [None] * len(delta.manifest.entries)
    )

    total_tables_size = 0
    for file_idx, manifest in enumerate(delta.manifest.entries):
        entry_pyarrow_table: LocalTable = (
            deltacat_storage.download_delta_manifest_entry(
                delta,
                file_idx,
                TableType.PYARROW,
                columns_to_compute,
                **deltacat_storage_kwargs,
            )
        )
        assert isinstance(entry_pyarrow_table, pyarrow.Table), (
            f"Stats collection is only supported for PyArrow tables, but received a table of "
            f"type '{type(entry_pyarrow_table)}' for manifest entry {file_idx} of delta: {delta.locator}."
        )
        total_tables_size += entry_pyarrow_table.nbytes
        if not columns_to_compute:
            columns_to_compute = entry_pyarrow_table.column_names

        for column_idx, pyarrow_column in enumerate(entry_pyarrow_table.columns):
            column_name = columns_to_compute[column_idx]
            column_stats_map[column_name][file_idx] = StatsResult.of(
                len(pyarrow_column), pyarrow_column.nbytes
            )

    # Add column-wide stats for a list of tables, these will be used for caching and retrieving later
    delta_ds_column_stats: List[DeltaColumnStats] = _to_dataset_column_stats(
        delta.locator, columns_to_compute, column_stats_map
    )

    dataset_stats: DeltaStats = DeltaStats.of(delta_ds_column_stats)

    # Quick validation for calculations
    assert dataset_stats.stats.pyarrow_table_bytes == total_tables_size, (
        f"Expected the size of all PyArrow tables ({total_tables_size} bytes) "
        f"to match the sum of each of its columns ({dataset_stats.stats.pyarrow_table_bytes} bytes)"
    )

    return dataset_stats


def _to_dataset_column_stats(
    delta_locator: DeltaLocator,
    column_names: List[str],
    column_manifest_map: Dict[str, List[Optional[StatsResult]]],
) -> List[DeltaColumnStats]:
    dataset_stats: List[DeltaColumnStats] = []
    for column_name in column_names:
        column_manifest_stats = ManifestEntryStats.of(
            column_manifest_map[column_name], delta_locator
        )
        dataset_column_stats = DeltaColumnStats.of(column_name, column_manifest_stats)
        dataset_stats.append(dataset_column_stats)
    return dataset_stats



import importlib
import logging
import time
from contextlib import nullcontext
from itertools import chain
from typing import Any, Dict, Generator, List, Optional, Tuple
import numpy as np
import pyarrow as pa
import ray
from deltacat import logs
from deltacat.compute.compactor import (
    DeltaAnnotated,
    DeltaFileEnvelope,
    RoundCompletionInfo,
)
from deltacat.storage.model.sort_key import SortKey
from deltacat.compute.compactor.model.delta_file_envelope import DeltaFileEnvelopeGroups
from deltacat.compute.compactor.model.hash_bucket_result import HashBucketResult
from deltacat.compute.compactor.utils import system_columns as sc
from deltacat.compute.compactor.utils.primary_key_index import (
    group_hash_bucket_indices,
    group_record_indices_by_hash_bucket,
)
from deltacat.storage import interface as unimplemented_deltacat_storage
from deltacat.types.media import StorageType
from deltacat.utils.common import sha1_digest
from deltacat.utils.ray_utils.runtime import (
    get_current_ray_task_id,
    get_current_ray_worker_id,
)
from deltacat.utils.common import ReadKwargsProvider
from deltacat.utils.performance import timed_invocation
from deltacat.utils.metrics import emit_timer_metrics, MetricsConfig
from deltacat.io.object_store import IObjectStore
from deltacat.utils.resources import get_current_node_peak_memory_usage_in_bytes

if importlib.util.find_spec("memray"):
    import memray

logger = logs.configure_deltacat_logger(logging.getLogger(__name__))

_PK_BYTES_DELIMITER = b"L6kl7u5f"


def _group_by_pk_hash_bucket(
    table: pa.Table, num_buckets: int, primary_keys: List[str]
) -> np.ndarray:
    # generate the primary key digest column
    all_pk_column_fields = []
    for pk_name in primary_keys:
        # casting a primary key column to numpy also ensures no nulls exist
        # TODO (pdames): catch error in cast to numpy and print user-friendly err msg.
        column_fields = table[pk_name].to_numpy()
        all_pk_column_fields.append(column_fields)
    hash_column_generator = _hash_pk_bytes_generator(all_pk_column_fields)
    table = sc.append_pk_hash_column(table, hash_column_generator)

    # drop primary key columns to free up memory
    table = table.drop(primary_keys)

    # group hash bucket record indices
    hash_bucket_to_indices = group_record_indices_by_hash_bucket(
        table,
        num_buckets,
    )

    # generate the ordered record number column
    hash_bucket_to_table = np.empty([num_buckets], dtype="object")
    for hb, indices in enumerate(hash_bucket_to_indices):
        if indices:
            hash_bucket_to_table[hb] = sc.append_record_idx_col(
                table.take(indices),
                indices,
            )
    return hash_bucket_to_table


def _hash_pk_bytes_generator(all_column_fields) -> Generator[bytes, None, None]:
    for field_index in range(len(all_column_fields[0])):
        bytes_to_join = []
        for column_fields in all_column_fields:
            bytes_to_join.append(bytes(str(column_fields[field_index]), "utf-8"))
        yield sha1_digest(_PK_BYTES_DELIMITER.join(bytes_to_join))


def _group_file_records_by_pk_hash_bucket(
    annotated_delta: DeltaAnnotated,
    num_hash_buckets: int,
    primary_keys: List[str],
    sort_key_names: List[str],
    is_src_delta: np.bool_ = True,
    read_kwargs_provider: Optional[ReadKwargsProvider] = None,
    deltacat_storage=unimplemented_deltacat_storage,
    deltacat_storage_kwargs: Optional[Dict[str, Any]] = None,
    **kwargs,
) -> Tuple[Optional[DeltaFileEnvelopeGroups], int]:
    if deltacat_storage_kwargs is None:
        deltacat_storage_kwargs = {}
    # read input parquet s3 objects into a list of delta file envelopes
    delta_file_envelopes, total_record_count = _read_delta_file_envelopes(
        annotated_delta,
        primary_keys,
        sort_key_names,
        read_kwargs_provider,
        deltacat_storage,
        deltacat_storage_kwargs,
        **kwargs,
    )
    if delta_file_envelopes is None:
        return None, 0

    # group the data by primary key hash value
    hb_to_delta_file_envelopes = np.empty([num_hash_buckets], dtype="object")
    for dfe in delta_file_envelopes:
        hash_bucket_to_table = _group_by_pk_hash_bucket(
            dfe.table,
            num_hash_buckets,
            primary_keys,
        )
        for hb, table in enumerate(hash_bucket_to_table):
            if table:
                if hb_to_delta_file_envelopes[hb] is None:
                    hb_to_delta_file_envelopes[hb] = []
                hb_to_delta_file_envelopes[hb].append(
                    DeltaFileEnvelope.of(
                        stream_position=dfe.stream_position,
                        file_index=dfe.file_index,
                        delta_type=dfe.delta_type,
                        table=table,
                        is_src_delta=is_src_delta,
                        file_record_count=dfe.file_record_count,
                    )
                )
    return hb_to_delta_file_envelopes, total_record_count


def _read_delta_file_envelopes(
    annotated_delta: DeltaAnnotated,
    primary_keys: List[str],
    sort_key_names: List[str],
    read_kwargs_provider: Optional[ReadKwargsProvider],
    deltacat_storage=unimplemented_deltacat_storage,
    deltacat_storage_kwargs: Optional[Dict[str, Any]] = None,
    **kwargs,
) -> Tuple[Optional[List[DeltaFileEnvelope]], int]:
    if deltacat_storage_kwargs is None:
        deltacat_storage_kwargs = {}
    columns_to_read = list(chain(primary_keys, sort_key_names))
    # TODO (rootliu) compare performance of column read from unpartitioned vs partitioned file
    # https://arrow.apache.org/docs/python/parquet.html#writing-to-partitioned-datasets
    tables = deltacat_storage.download_delta(
        annotated_delta,
        max_parallelism=1,
        columns=columns_to_read,
        file_reader_kwargs_provider=read_kwargs_provider,
        storage_type=StorageType.LOCAL,
        **deltacat_storage_kwargs,
    )
    annotations = annotated_delta.annotations
    assert (
        len(tables) == len(annotations),
        f"Unexpected Error: Length of downloaded delta manifest tables "
        f"({len(tables)}) doesn't match the length of delta manifest "
        f"annotations ({len(annotations)}).",
    )
    if not tables:
        return None, 0

    delta_file_envelopes = []
    total_record_count = 0
    for i, table in enumerate(tables):
        total_record_count += len(table)
        delta_file = DeltaFileEnvelope.of(
            stream_position=annotations[i].annotation_stream_position,
            file_index=annotations[i].annotation_file_index,
            delta_type=annotations[i].annotation_delta_type,
            table=table,
            file_record_count=len(table),
        )
        delta_file_envelopes.append(delta_file)
    return delta_file_envelopes, total_record_count


def _timed_hash_bucket(
    annotated_delta: DeltaAnnotated,
    round_completion_info: Optional[RoundCompletionInfo],
    primary_keys: List[str],
    sort_keys: List[SortKey],
    num_buckets: int,
    num_groups: int,
    enable_profiler: bool,
    read_kwargs_provider: Optional[ReadKwargsProvider] = None,
    object_store: Optional[IObjectStore] = None,
    deltacat_storage=unimplemented_deltacat_storage,
    deltacat_storage_kwargs: Optional[Dict[str, Any]] = None,
    **kwargs,
):
    if deltacat_storage_kwargs is None:
        deltacat_storage_kwargs = {}
    task_id = get_current_ray_task_id()
    worker_id = get_current_ray_worker_id()
    with memray.Tracker(
        f"hash_bucket_{worker_id}_{task_id}.bin"
    ) if enable_profiler else nullcontext():
        sort_key_names = [key.key_name for key in sort_keys]
        if not round_completion_info:
            is_src_delta = True
        else:
            is_src_delta = (
                annotated_delta.locator.partition_locator
                != round_completion_info.compacted_delta_locator.partition_locator
            )
        (
            delta_file_envelope_groups,
            total_record_count,
        ) = _group_file_records_by_pk_hash_bucket(
            annotated_delta,
            num_buckets,
            primary_keys,
            sort_key_names,
            is_src_delta,
            read_kwargs_provider,
            deltacat_storage,
            deltacat_storage_kwargs,
            **kwargs,
        )
        hash_bucket_group_to_obj_id, _ = group_hash_bucket_indices(
            delta_file_envelope_groups, num_buckets, num_groups, object_store
        )

        peak_memory_usage_bytes = get_current_node_peak_memory_usage_in_bytes()
        return HashBucketResult(
            hash_bucket_group_to_obj_id,
            np.int64(total_record_count),
            np.double(peak_memory_usage_bytes),
            np.double(0.0),
            np.double(time.time()),
        )


@ray.remote
def hash_bucket(
    annotated_delta: DeltaAnnotated,
    round_completion_info: Optional[RoundCompletionInfo],
    primary_keys: List[str],
    sort_keys: List[SortKey],
    num_buckets: int,
    num_groups: int,
    enable_profiler: bool,
    metrics_config: MetricsConfig,
    read_kwargs_provider: Optional[ReadKwargsProvider],
    object_store: Optional[IObjectStore],
    deltacat_storage=unimplemented_deltacat_storage,
    deltacat_storage_kwargs: Optional[Dict[str, Any]] = None,
    **kwargs,
) -> HashBucketResult:
    if deltacat_storage_kwargs is None:
        deltacat_storage_kwargs = {}
    logger.info(f"Starting hash bucket task...")
    hash_bucket_result, duration = timed_invocation(
        func=_timed_hash_bucket,
        annotated_delta=annotated_delta,
        round_completion_info=round_completion_info,
        primary_keys=primary_keys,
        sort_keys=sort_keys,
        num_buckets=num_buckets,
        num_groups=num_groups,
        enable_profiler=enable_profiler,
        read_kwargs_provider=read_kwargs_provider,
        object_store=object_store,
        deltacat_storage=deltacat_storage,
        deltacat_storage_kwargs=deltacat_storage_kwargs,
        **kwargs,
    )

    emit_metrics_time = 0.0
    if metrics_config:
        emit_result, latency = timed_invocation(
            func=emit_timer_metrics,
            metrics_name="hash_bucket",
            value=duration,
            metrics_config=metrics_config,
        )
        emit_metrics_time = latency

    logger.info(f"Finished hash bucket task...")
    return HashBucketResult(
        hash_bucket_result[0],
        hash_bucket_result[1],
        hash_bucket_result[2],
        np.double(emit_metrics_time),
        hash_bucket_result[4],
    )



import logging
from typing import Optional, List

from daft.table import read_parquet_into_pyarrow
from daft import TimeUnit
from daft.io import IOConfig, S3Config
import pyarrow as pa

from deltacat import logs
from deltacat.utils.common import ReadKwargsProvider
from deltacat.utils.schema import coerce_pyarrow_table_to_schema

from deltacat.types.media import ContentType, ContentEncoding
from deltacat.aws.constants import BOTO_MAX_RETRIES, DAFT_MAX_S3_CONNECTIONS_PER_FILE
from deltacat.utils.performance import timed_invocation

from deltacat.types.partial_download import (
    PartialFileDownloadParams,
)


logger = logs.configure_deltacat_logger(logging.getLogger(__name__))


def daft_s3_file_to_table(
    s3_url: str,
    content_type: str,
    content_encoding: str,
    column_names: Optional[List[str]] = None,
    include_columns: Optional[List[str]] = None,
    pa_read_func_kwargs_provider: Optional[ReadKwargsProvider] = None,
    partial_file_download_params: Optional[PartialFileDownloadParams] = None,
    **s3_client_kwargs,
):
    assert (
        content_type == ContentType.PARQUET.value
    ), f"daft native reader currently only supports parquet, got {content_type}"

    assert (
        content_encoding == ContentEncoding.IDENTITY.value
    ), f"daft native reader currently only supports identity encoding, got {content_encoding}"

    kwargs = {}
    if pa_read_func_kwargs_provider is not None:
        kwargs = pa_read_func_kwargs_provider(content_type, kwargs)

    coerce_int96_timestamp_unit = TimeUnit.from_str(
        kwargs.get("coerce_int96_timestamp_unit", "ms")
    )

    row_groups = None
    if (
        partial_file_download_params is not None
        and partial_file_download_params.row_groups_to_download is not None
    ):
        row_groups = partial_file_download_params.row_groups_to_download

    io_config = IOConfig(
        s3=S3Config(
            key_id=s3_client_kwargs.get("aws_access_key_id"),
            access_key=s3_client_kwargs.get("aws_secret_access_key"),
            session_token=s3_client_kwargs.get("aws_session_token"),
            retry_mode="adaptive",
            num_tries=BOTO_MAX_RETRIES,
            max_connections=DAFT_MAX_S3_CONNECTIONS_PER_FILE,
        )
    )

    pa_table, latency = timed_invocation(
        read_parquet_into_pyarrow,
        path=s3_url,
        columns=include_columns or column_names,
        row_groups=row_groups,
        io_config=io_config,
        coerce_int96_timestamp_unit=coerce_int96_timestamp_unit,
        multithreaded_io=False,
    )

    logger.debug(f"Time to read S3 object from {s3_url} into daft table: {latency}s")

    if kwargs.get("schema") is not None:
        input_schema = kwargs["schema"]
        if include_columns is not None:
            input_schema = pa.schema(
                [input_schema.field(col) for col in include_columns],
                metadata=input_schema.metadata,
            )
        elif column_names is not None:
            input_schema = pa.schema(
                [input_schema.field(col) for col in column_names],
                metadata=input_schema.metadata,
            )
        return coerce_pyarrow_table_to_schema(pa_table, input_schema)
    else:
        return pa_table



# Allow classes to use self-referencing Type hints in Python 3.7.
from __future__ import annotations

from enum import Enum


class SortOrder(str, Enum):
    ASCENDING = "ascending"
    DESCENDING = "descending"


class SortKey(tuple):
    @staticmethod
    def of(key_name: str, sort_order: SortOrder = SortOrder.ASCENDING) -> SortKey:
        """
        Create a sort key from a field name to use as the sort key, and
        the sort order for this key. If no sort order is specified, then the
        data will be sorted in ascending order by default. Note that compaction
        always keeps the LAST occurrence of this key post-sort. For example, if
        you used an integer column as your sort key which contained the values
        [2, 1, 3] specifying SortOrder.ASCENDING would ensure that the
        value [3] is kept over [2, 1], and specifying SortOrder.DESCENDING
        would ensure that [1] is kept over [2, 3].
        """
        return SortKey((key_name, sort_order.value))

    @property
    def key_name(self) -> str:
        return self[0]

    @property
    def sort_order(self) -> SortOrder:
        return SortOrder(self[1])



from enum import Enum
from typing import Set


class StatsType(str, Enum):
    ROW_COUNT = "rowCount"
    PYARROW_TABLE_BYTES = "pyarrowTableBytes"


ALL_STATS_TYPES: Set[StatsType] = {field for field in StatsType}



from typing import Any, Callable, Dict, List, Optional, Set, Union, Tuple

import pyarrow as pa
import json
import sqlite3
from sqlite3 import Cursor, Connection
import uuid
import io

from deltacat.tests.test_utils.storage import create_empty_delta
from deltacat.utils.common import current_time_ms

from deltacat.storage import (
    Delta,
    DeltaLocator,
    DeltaType,
    DistributedDataset,
    LifecycleState,
    ListResult,
    LocalDataset,
    LocalTable,
    Manifest,
    ManifestAuthor,
    Namespace,
    NamespaceLocator,
    Partition,
    SchemaConsistencyType,
    Stream,
    StreamLocator,
    Table,
    TableVersion,
    TableVersionLocator,
    TableLocator,
    CommitState,
    SortKey,
    PartitionLocator,
    ManifestMeta,
    ManifestEntry,
    ManifestEntryList,
)
from deltacat.types.media import ContentType, StorageType, TableType, ContentEncoding
from deltacat.utils.common import ReadKwargsProvider

SQLITE_CUR_ARG = "sqlite3_cur"
SQLITE_CON_ARG = "sqlite3_con"
DB_FILE_PATH_ARG = "db_file_path"

STORAGE_TYPE = "SQLITE3"
STREAM_ID_PROPERTY = "stream_id"
CREATE_NAMESPACES_TABLE = (
    "CREATE TABLE IF NOT EXISTS namespaces(locator, value, PRIMARY KEY (locator))"
)
CREATE_TABLES_TABLE = (
    "CREATE TABLE IF NOT EXISTS tables(locator, namespace_locator, value, PRIMARY KEY (locator), "
    "FOREIGN KEY (namespace_locator) REFERENCES namespaces(locator))"
)
CREATE_TABLE_VERSIONS_TABLE = (
    "CREATE TABLE IF NOT EXISTS table_versions(locator, table_locator, value, PRIMARY KEY (locator), "
    "FOREIGN KEY (table_locator) REFERENCES tables(locator))"
)
CREATE_STREAMS_TABLE = (
    "CREATE TABLE IF NOT EXISTS streams(locator, table_version_locator, value, PRIMARY KEY(locator), "
    "FOREIGN KEY (table_version_locator) REFERENCES table_versions(locator))"
)
CREATE_PARTITIONS_TABLE = (
    "CREATE TABLE IF NOT EXISTS partitions(locator, stream_locator, value, PRIMARY KEY(locator), "
    "FOREIGN KEY (stream_locator) REFERENCES streams(locator))"
)
CREATE_DELTAS_TABLE = (
    "CREATE TABLE IF NOT EXISTS deltas(locator, partition_locator, value, PRIMARY KEY(locator), "
    "FOREIGN KEY (partition_locator) REFERENCES partitions(locator))"
)
CREATE_DATA_TABLE = "CREATE TABLE IF NOT EXISTS data(uri, value, PRIMARY KEY(uri))"


def _get_sqlite3_cursor_con(kwargs) -> Tuple[Cursor, Connection]:
    if SQLITE_CUR_ARG in kwargs and SQLITE_CON_ARG in kwargs:
        return kwargs[SQLITE_CUR_ARG], kwargs[SQLITE_CON_ARG]
    elif DB_FILE_PATH_ARG in kwargs:
        con = sqlite3.connect(kwargs[DB_FILE_PATH_ARG])
        cur = con.cursor()
        return cur, con

    raise ValueError(f"Invalid local db connection kwargs: {kwargs}")


def _get_manifest_entry_uri(manifest_entry_id: str) -> str:
    return f"cloudpickle://{manifest_entry_id}"


def list_namespaces(*args, **kwargs) -> ListResult[Namespace]:
    cur, con = _get_sqlite3_cursor_con(kwargs)
    res = cur.execute("SELECT * FROM namespaces")
    fetched = res.fetchall()
    result = []

    for item in fetched:
        result.append(Namespace(json.loads(item[1])))

    return ListResult.of(result, None, None)


def list_tables(namespace: str, *args, **kwargs) -> ListResult[Table]:
    cur, con = _get_sqlite3_cursor_con(kwargs)
    params = (NamespaceLocator.of(namespace).canonical_string(),)
    res = cur.execute("SELECT * FROM tables WHERE namespace_locator = ?", params)
    fetched = res.fetchall()
    result = []

    for item in fetched:
        result.append(Table(json.loads(item[2])))

    return ListResult.of(result, None, None)


def list_table_versions(
    namespace: str, table_name: str, *args, **kwargs
) -> ListResult[TableVersion]:
    cur, con = _get_sqlite3_cursor_con(kwargs)
    table_locator = TableLocator.of(NamespaceLocator.of(namespace), table_name)

    res = cur.execute(
        "SELECT * FROM table_versions WHERE table_locator = ?",
        (table_locator.canonical_string(),),
    )
    fetched = res.fetchall()
    result = []

    for item in fetched:
        result.append(TableVersion(json.loads(item[2])))

    return ListResult.of(result, None, None)


def list_partitions(
    namespace: str,
    table_name: str,
    table_version: Optional[str] = None,
    *args,
    **kwargs,
) -> ListResult[Partition]:
    cur, con = _get_sqlite3_cursor_con(kwargs)

    stream = get_stream(namespace, table_name, table_version, *args, **kwargs)

    res = cur.execute(
        "SELECT * FROM partitions WHERE stream_locator = ?",
        (stream.locator.canonical_string(),),
    )

    fetched = res.fetchall()
    result = []
    for item in fetched:
        partition = Partition(json.loads(item[2]))
        if partition.state == CommitState.COMMITTED:
            result.append(partition)

    return ListResult.of(result, None, None)


def list_stream_partitions(stream: Stream, *args, **kwargs) -> ListResult[Partition]:
    return list_partitions(
        stream.namespace, stream.table_name, stream.table_version, *args, **kwargs
    )


def list_deltas(
    namespace: str,
    table_name: str,
    partition_values: Optional[List[Any]] = None,
    table_version: Optional[str] = None,
    first_stream_position: Optional[int] = None,
    last_stream_position: Optional[int] = None,
    ascending_order: Optional[bool] = None,
    include_manifest: bool = False,
    *args,
    **kwargs,
) -> ListResult[Delta]:
    stream = get_stream(namespace, table_name, table_version, *args, **kwargs)
    if stream is None:
        return ListResult.of([], None, None)

    partition = get_partition(stream.locator, partition_values, *args, **kwargs)

    all_deltas = list_partition_deltas(
        partition,
        first_stream_position=first_stream_position,
        last_stream_position=last_stream_position,
        ascending_order=ascending_order,
        include_manifest=include_manifest,
        *args,
        **kwargs,
    ).all_items()

    result = []

    for delta in all_deltas:
        if (
            not first_stream_position or first_stream_position < delta.stream_position
        ) and (
            not last_stream_position or delta.stream_position <= last_stream_position
        ):
            result.append(delta)

        if not include_manifest:
            delta.manifest = None

    result.sort(reverse=(not ascending_order), key=lambda d: d.stream_position)
    return ListResult.of(result, None, None)


def list_partition_deltas(
    partition_like: Union[Partition, PartitionLocator],
    first_stream_position: Optional[int] = None,
    last_stream_position: Optional[int] = None,
    ascending_order: bool = False,
    include_manifest: bool = False,
    *args,
    **kwargs,
) -> ListResult[Delta]:
    cur, con = _get_sqlite3_cursor_con(kwargs)

    if partition_like is None:
        return ListResult.of([], None, None)

    if first_stream_position is None:
        first_stream_position = 0

    if last_stream_position is None:
        last_stream_position = float("inf")

    assert isinstance(partition_like, Partition) or isinstance(
        partition_like, PartitionLocator
    ), f"Expected a Partition or PartitionLocator as an input argument but found {partition_like}"

    partition_locator = None
    if isinstance(partition_like, Partition):
        partition_locator = partition_like.locator
    else:
        partition_locator = partition_like

    res = cur.execute(
        "SELECT * FROM deltas WHERE partition_locator = ?",
        (partition_locator.canonical_string(),),
    )

    serialized_items = res.fetchall()

    if not serialized_items:
        return ListResult.of([], None, None)

    result = []
    for item in serialized_items:
        current_delta = Delta(json.loads(item[2]))
        if (
            first_stream_position
            <= current_delta.stream_position
            <= last_stream_position
        ):
            result.append(current_delta)

        if not include_manifest:
            current_delta.manifest = None

    result.sort(
        reverse=True if not ascending_order else False, key=lambda d: d.stream_position
    )
    return ListResult.of(result, None, None)


def get_delta(
    namespace: str,
    table_name: str,
    stream_position: int,
    partition_values: Optional[List[Any]] = None,
    table_version: Optional[str] = None,
    include_manifest: bool = False,
    *args,
    **kwargs,
) -> Optional[Delta]:
    cur, con = _get_sqlite3_cursor_con(kwargs)

    stream = get_stream(namespace, table_name, table_version, *args, **kwargs)
    partition = get_partition(stream.locator, partition_values, *args, **kwargs)
    delta_locator = DeltaLocator.of(partition.locator, stream_position)

    res = cur.execute(
        "SELECT * FROM deltas WHERE locator = ?", (delta_locator.canonical_string(),)
    )

    serialized_delta = res.fetchone()
    if serialized_delta is None:
        return None

    delta = Delta(json.loads(serialized_delta[2]))

    if not include_manifest:
        delta.manifest = None

    return delta


def get_latest_delta(
    namespace: str,
    table_name: str,
    partition_values: Optional[List[Any]] = None,
    table_version: Optional[str] = None,
    include_manifest: bool = False,
    *args,
    **kwargs,
) -> Optional[Delta]:

    deltas = list_deltas(
        namespace,
        table_name,
        partition_values,
        table_version,
        None,
        None,
        False,
        include_manifest,
        *args,
        **kwargs,
    ).all_items()

    if not deltas:
        return None

    return deltas[0]


def download_delta(
    delta_like: Union[Delta, DeltaLocator],
    table_type: TableType = TableType.PYARROW,
    storage_type: StorageType = StorageType.DISTRIBUTED,
    max_parallelism: Optional[int] = None,
    columns: Optional[List[str]] = None,
    file_reader_kwargs_provider: Optional[ReadKwargsProvider] = None,
    ray_options_provider: Callable[[int, Any], Dict[str, Any]] = None,
    *args,
    **kwargs,
) -> Union[LocalDataset, DistributedDataset]:
    result = []
    manifest = get_delta_manifest(delta_like, *args, **kwargs)

    for entry_index in range(len(manifest.entries)):
        result.append(
            download_delta_manifest_entry(
                delta_like=delta_like,
                entry_index=entry_index,
                table_type=table_type,
                columns=columns,
                file_reader_kwargs_provider=file_reader_kwargs_provider,
                *args,
                **kwargs,
            )
        )

    return result


def download_delta_manifest_entry(
    delta_like: Union[Delta, DeltaLocator],
    entry_index: int,
    table_type: TableType = TableType.PYARROW,
    columns: Optional[List[str]] = None,
    file_reader_kwargs_provider: Optional[ReadKwargsProvider] = None,
    *args,
    **kwargs,
) -> LocalTable:
    cur, con = _get_sqlite3_cursor_con(kwargs)

    manifest = get_delta_manifest(delta_like, *args, **kwargs)
    if entry_index >= len(manifest.entries):
        raise IndexError(
            f"Manifest entry index {entry_index} does not exist. "
            f"Valid values: [0, {len(manifest.entries)}]"
        )

    entry = manifest.entries[entry_index]

    res = cur.execute("SELECT value FROM data WHERE uri = ?", (entry.uri,))
    serialized_data = res.fetchone()

    if serialized_data is None:
        raise ValueError(
            f"Invalid value of delta locator: {delta_like.canonical_string()}"
        )

    serialized_data = serialized_data[0]
    if entry.meta.content_type == ContentType.PARQUET:
        if table_type == TableType.PYARROW_PARQUET:
            table = pa.parquet.ParquetFile(io.BytesIO(serialized_data))
        else:
            table = pa.parquet.read_table(io.BytesIO(serialized_data), columns=columns)
    elif entry.meta.content_type == ContentType.UNESCAPED_TSV:
        assert (
            table_type != TableType.PYARROW_PARQUET
        ), f"uTSV table cannot be read as {table_type}"
        parse_options = pa.csv.ParseOptions(delimiter="\t")
        convert_options = pa.csv.ConvertOptions(
            null_values=[""], strings_can_be_null=True, include_columns=columns
        )
        table = pa.csv.read_csv(
            io.BytesIO(serialized_data),
            parse_options=parse_options,
            convert_options=convert_options,
        )
    else:
        raise ValueError(f"Content type: {entry.meta.content_type} not supported.")

    if table_type == TableType.PYARROW:
        return table
    elif table_type == TableType.PYARROW_PARQUET:
        return table
    elif table_type == TableType.NUMPY:
        raise NotImplementedError(f"Table type={table_type} not supported")
    elif table_type == TableType.PANDAS:
        return table.to_pandas()

    return table


def get_delta_manifest(
    delta_like: Union[Delta, DeltaLocator], *args, **kwargs
) -> Optional[Manifest]:
    delta = get_delta(
        namespace=delta_like.namespace,
        table_name=delta_like.table_name,
        stream_position=delta_like.stream_position,
        partition_values=delta_like.partition_values,
        table_version=delta_like.table_version,
        include_manifest=True,
        *args,
        **kwargs,
    )
    if not delta:
        return None

    return delta.manifest


def create_namespace(
    namespace: str, permissions: Dict[str, Any], *args, **kwargs
) -> Namespace:
    cur, con = _get_sqlite3_cursor_con(kwargs)
    locator = NamespaceLocator.of(namespace)
    result = Namespace.of(locator, permissions)
    params = (locator.canonical_string(), json.dumps(result))
    cur.execute(CREATE_NAMESPACES_TABLE)
    cur.execute(CREATE_TABLES_TABLE)
    cur.execute(CREATE_TABLE_VERSIONS_TABLE)
    cur.execute(CREATE_STREAMS_TABLE)
    cur.execute(CREATE_PARTITIONS_TABLE)
    cur.execute(CREATE_DELTAS_TABLE)
    cur.execute(CREATE_DATA_TABLE)
    cur.execute("INSERT OR IGNORE INTO namespaces VALUES(?, ?)", params)
    con.commit()
    return result


def update_namespace(
    namespace: str,
    permissions: Optional[Dict[str, Any]] = None,
    new_namespace: Optional[str] = None,
    *args,
    **kwargs,
) -> None:
    assert new_namespace is None, "namespace name cannot be changed"
    cur, con = _get_sqlite3_cursor_con(kwargs)
    locator = NamespaceLocator.of(namespace)
    result = Namespace.of(locator, permissions)
    params = (json.dumps(result), locator.canonical_string())
    cur.execute("UPDATE namespaces SET value = ? WHERE locator = ?", params)
    con.commit()


def create_table_version(
    namespace: str,
    table_name: str,
    table_version: Optional[str] = None,
    schema: Optional[Union[pa.Schema, str, bytes]] = None,
    schema_consistency: Optional[Dict[str, SchemaConsistencyType]] = None,
    partition_keys: Optional[List[Dict[str, Any]]] = None,
    primary_key_column_names: Optional[Set[str]] = None,
    sort_keys: Optional[List[SortKey]] = None,
    table_version_description: Optional[str] = None,
    table_version_properties: Optional[Dict[str, str]] = None,
    table_permissions: Optional[Dict[str, Any]] = None,
    table_description: Optional[str] = None,
    table_properties: Optional[Dict[str, str]] = None,
    supported_content_types: Optional[List[ContentType]] = None,
    *args,
    **kwargs,
) -> Stream:
    cur, con = _get_sqlite3_cursor_con(kwargs)

    latest_version = get_latest_table_version(namespace, table_name, *args, **kwargs)
    if (
        table_version is not None
        and latest_version
        and int(latest_version.table_version) + 1 != int(table_version)
    ):
        raise AssertionError(
            f"Table version can only be incremented. Last version={latest_version.table_version}"
        )
    elif table_version is None:
        table_version = (
            (int(latest_version.table_version) + 1) if latest_version else "1"
        )

    table_locator = TableLocator.of(NamespaceLocator.of(namespace), table_name)
    table_obj = Table.of(
        table_locator, table_permissions, table_description, table_properties
    )
    table_version_locator = TableVersionLocator.of(
        table_locator=table_locator, table_version=table_version
    )

    stream_id = uuid.uuid4().__str__()

    if table_version_properties is None:
        table_version_properties = {}

    properties = {**table_version_properties, STREAM_ID_PROPERTY: stream_id}
    table_version_obj = TableVersion.of(
        table_version_locator,
        schema=schema,
        partition_keys=partition_keys,
        primary_key_columns=primary_key_column_names,
        description=table_version_description,
        properties=properties,
        sort_keys=sort_keys,
        content_types=supported_content_types,
    )
    stream_locator = StreamLocator.of(
        table_version_obj.locator, stream_id=stream_id, storage_type=STORAGE_TYPE
    )
    result_stream = Stream.of(
        stream_locator, partition_keys=partition_keys, state=CommitState.COMMITTED
    )

    params = (
        table_locator.canonical_string(),
        table_locator.namespace_locator.canonical_string(),
        json.dumps(table_obj),
    )
    cur.execute("INSERT OR IGNORE INTO tables VALUES (?, ?, ?)", params)
    params = (
        table_version_locator.canonical_string(),
        table_locator.canonical_string(),
        json.dumps(table_version_obj),
    )
    cur.execute("INSERT OR IGNORE INTO table_versions VALUES (?, ?, ?)", params)

    params = (
        stream_locator.canonical_string(),
        table_version_locator.canonical_string(),
        json.dumps(result_stream),
    )
    cur.execute("INSERT OR IGNORE INTO streams VALUES (?, ?, ?)", params)
    con.commit()
    return result_stream


def update_table(
    namespace: str,
    table_name: str,
    permissions: Optional[Dict[str, Any]] = None,
    description: Optional[str] = None,
    properties: Optional[Dict[str, str]] = None,
    new_table_name: Optional[str] = None,
    *args,
    **kwargs,
) -> None:
    cur, con = _get_sqlite3_cursor_con(kwargs)
    table_locator = TableLocator.of(NamespaceLocator.of(namespace), table_name)
    table_obj = Table.of(table_locator, permissions, description, properties)

    params = (table_locator.canonical_string(),)
    cur.execute("DELETE FROM tables WHERE locator = ?", params)
    params = (
        table_locator.canonical_string(),
        table_locator.namespace_locator.canonical_string(),
        json.dumps(table_obj),
    )
    cur.execute("INSERT INTO tables VALUES (?, ?, ?)", params)
    con.commit()


def update_table_version(
    namespace: str,
    table_name: str,
    table_version: str,
    lifecycle_state: Optional[LifecycleState] = None,
    schema: Optional[Union[pa.Schema, str, bytes]] = None,
    schema_consistency: Optional[Dict[str, SchemaConsistencyType]] = None,
    description: Optional[str] = None,
    properties: Optional[Dict[str, str]] = None,
    *args,
    **kwargs,
) -> None:
    cur, con = _get_sqlite3_cursor_con(kwargs)
    table_locator = TableLocator.of(NamespaceLocator.of(namespace), table_name)
    table_version_locator = TableVersionLocator.of(
        table_locator=table_locator, table_version=table_version
    )

    res = cur.execute(
        "SELECT * from table_versions WHERE locator = ?",
        (table_version_locator.canonical_string(),),
    )
    serialized_table_version = res.fetchone()
    assert (
        serialized_table_version is not None
    ), f"Table version not found with locator={table_version_locator.canonical_string()}"
    current_table_version_obj = TableVersion(json.loads(serialized_table_version[2]))

    if properties is None:
        properties = {}

    current_props = (
        current_table_version_obj.properties
        if current_table_version_obj.properties
        else {}
    )

    tv_properties = {**properties, **current_props}
    table_version_obj = TableVersion.of(
        table_version_locator,
        schema=schema,
        partition_keys=current_table_version_obj.partition_keys,
        primary_key_columns=current_table_version_obj.primary_keys,
        description=description,
        properties=tv_properties,
        sort_keys=current_table_version_obj.sort_keys,
        content_types=current_table_version_obj.content_types,
    )

    params = (
        table_locator.canonical_string(),
        json.dumps(table_version_obj),
        table_version_locator.canonical_string(),
    )
    cur.execute(
        "UPDATE table_versions SET table_locator = ?, value = ? WHERE locator = ?",
        params,
    )
    con.commit()


def stage_stream(
    namespace: str,
    table_name: str,
    table_version: Optional[str] = None,
    *args,
    **kwargs,
) -> Stream:
    cur, con = _get_sqlite3_cursor_con(kwargs)

    existing_table_version = get_table_version(
        namespace, table_name, table_version, *args, **kwargs
    )
    existing_stream = get_stream(namespace, table_name, table_version, *args, **kwargs)

    stream_id = uuid.uuid4().__str__()
    new_stream_locator = StreamLocator.of(
        existing_table_version.locator, stream_id, STORAGE_TYPE
    )
    new_stream = Stream.of(
        new_stream_locator,
        existing_stream.partition_keys,
        CommitState.STAGED,
        existing_stream.locator.canonical_string(),
    )

    params = (
        new_stream_locator.canonical_string(),
        existing_table_version.locator.canonical_string(),
        json.dumps(new_stream),
    )
    cur.execute("INSERT INTO streams VALUES (?, ?, ?)", params)
    con.commit()

    return new_stream


def commit_stream(stream: Stream, *args, **kwargs) -> Stream:
    cur, con = _get_sqlite3_cursor_con(kwargs)

    existing_table_version = get_table_version(
        stream.namespace, stream.table_name, stream.table_version, *args, **kwargs
    )
    stream_to_commit = Stream.of(
        stream.locator,
        stream.partition_keys,
        CommitState.COMMITTED,
        stream.previous_stream_digest,
    )

    existing_table_version.properties[
        STREAM_ID_PROPERTY
    ] = stream_to_commit.locator.stream_id

    params = (
        json.dumps(existing_table_version),
        existing_table_version.locator.canonical_string(),
    )
    cur.execute("UPDATE table_versions SET value = ? WHERE locator = ?", params)
    params = (json.dumps(stream_to_commit), stream_to_commit.locator.canonical_string())
    cur.execute("UPDATE streams SET value = ? WHERE locator = ?", params)
    con.commit()

    return stream_to_commit


def delete_stream(
    namespace: str,
    table_name: str,
    table_version: Optional[str] = None,
    *args,
    **kwargs,
) -> None:
    cur, con = _get_sqlite3_cursor_con(kwargs)

    table_version_locator = TableVersionLocator.of(
        TableLocator.of(NamespaceLocator.of(namespace), table_name), table_version
    )

    res = cur.execute(
        "SELECT locator FROM streams WHERE table_version_locator = ?",
        (table_version_locator.canonical_string(),),
    )
    locators = res.fetchall()
    cur.executemany("DELETE FROM streams WHERE locator = ?", locators)
    cur.execute(
        "DELETE FROM table_versions WHERE locator = ?",
        (table_version_locator.canonical_string(),),
    )

    con.commit()


def stage_partition(
    stream: Stream, partition_values: Optional[List[Any]] = None, *args, **kwargs
) -> Partition:
    cur, con = _get_sqlite3_cursor_con(kwargs)
    partition_id = uuid.uuid4().__str__()
    partition_locator = PartitionLocator.of(
        stream.locator, partition_values=partition_values, partition_id=partition_id
    )

    tv = get_table_version(
        stream.namespace, stream.table_name, stream.table_version, *args, **kwargs
    )

    pv_partition = get_partition(
        stream.locator, partition_values=partition_values, *args, **kwargs
    )

    stream_position = current_time_ms()
    partition = Partition.of(
        partition_locator,
        schema=tv.schema,
        content_types=tv.content_types,
        state=CommitState.STAGED,
        previous_stream_position=pv_partition.stream_position if pv_partition else None,
        previous_partition_id=pv_partition.partition_id if pv_partition else None,
        stream_position=stream_position,
    )

    params = (
        partition.locator.canonical_string(),
        partition.stream_locator.canonical_string(),
        json.dumps(partition),
    )
    cur.execute("INSERT INTO partitions VALUES (?, ?, ?)", params)
    con.commit()

    return partition


def commit_partition(partition: Partition, *args, **kwargs) -> Partition:
    cur, con = _get_sqlite3_cursor_con(kwargs)
    pv_partition = get_partition(
        partition.stream_locator,
        partition_values=partition.partition_values,
        *args,
        **kwargs,
    )

    # deprecate old and commit new one
    if pv_partition:
        pv_partition.state = CommitState.DEPRECATED
        params = (json.dumps(pv_partition), pv_partition.locator.canonical_string())
        cur.execute("UPDATE partitions SET value = ? WHERE locator = ?", params)

    deltas = list_partition_deltas(partition, *args, **kwargs).all_items()
    deltas.sort(reverse=True, key=lambda x: x.stream_position)

    stream_position = partition.stream_position
    if deltas:
        stream_position = deltas[0].stream_position

    partition.state = CommitState.COMMITTED
    partition.stream_position = stream_position
    partition.previous_stream_position = (
        pv_partition.stream_position if pv_partition else None
    )
    params = (json.dumps(partition), partition.locator.canonical_string())
    cur.execute("UPDATE partitions SET value = ? WHERE locator = ?", params)
    con.commit()

    return partition


def delete_partition(
    namespace: str,
    table_name: str,
    table_version: Optional[str] = None,
    partition_values: Optional[List[Any]] = None,
    *args,
    **kwargs,
) -> None:
    cur, con = _get_sqlite3_cursor_con(kwargs)
    stream = get_stream(namespace, table_name, table_version, *args, **kwargs)
    partition = get_partition(stream.locator, partition_values, *args, **kwargs)

    partition.state = CommitState.DEPRECATED
    params = (json.dumps(partition), partition.locator.canonical_string())

    cur.execute("UPDATE partitions SET value = ? WHERE locator = ?", params)
    con.commit()


def get_partition(
    stream_locator: StreamLocator,
    partition_values: Optional[List[Any]] = None,
    *args,
    **kwargs,
) -> Optional[Partition]:
    cur, con = _get_sqlite3_cursor_con(kwargs)

    res = cur.execute(
        "SELECT * FROM partitions WHERE stream_locator = ?",
        (stream_locator.canonical_string(),),
    )

    serialized_partitions = res.fetchall()

    if not serialized_partitions:
        return None

    if partition_values is None:
        partition_values = []

    prior_pv = ",".join(partition_values)

    for item in serialized_partitions:
        partition = Partition(json.loads(item[2]))
        pv = ",".join(partition.partition_values if partition.partition_values else [])

        if pv == prior_pv and partition.state == CommitState.COMMITTED:
            return partition

    return None


def stage_delta(
    data: Union[LocalTable, LocalDataset, DistributedDataset, Manifest],
    partition: Partition,
    delta_type: DeltaType = DeltaType.UPSERT,
    max_records_per_entry: Optional[int] = None,
    author: Optional[ManifestAuthor] = None,
    properties: Optional[Dict[str, str]] = None,
    s3_table_writer_kwargs: Optional[Dict[str, Any]] = None,
    content_type: ContentType = ContentType.PARQUET,
    *args,
    **kwargs,
) -> Delta:
    cur, con = _get_sqlite3_cursor_con(kwargs)
    manifest_entry_id = uuid.uuid4().__str__()
    uri = _get_manifest_entry_uri(manifest_entry_id)

    if data is None:
        delta = create_empty_delta(
            partition,
            delta_type,
            author,
            properties=properties,
            manifest_entry_id=manifest_entry_id,
        )
        cur.execute("INSERT OR IGNORE INTO data VALUES (?, ?)", (uri, None))
        params = (delta.locator.canonical_string(), "staged_delta", json.dumps(delta))
        cur.execute("INSERT OR IGNORE INTO deltas VALUES (?, ?, ?)", params)
        con.commit()
        return delta

    serialized_data = None
    if content_type == ContentType.PARQUET:
        buffer = io.BytesIO()
        pa.parquet.write_table(data, buffer)
        serialized_data = buffer.getvalue()
    elif content_type == ContentType.UNESCAPED_TSV:
        buffer = io.BytesIO()
        write_options = pa.csv.WriteOptions(
            include_header=True, delimiter="\t", quoting_style="none"
        )
        pa.csv.write_csv(data, buffer, write_options=write_options)
        serialized_data = buffer.getvalue()
    else:
        raise ValueError(f"Unsupported content type: {content_type}")

    stream_position = current_time_ms()
    delta_locator = DeltaLocator.of(partition.locator, stream_position=stream_position)

    meta = ManifestMeta.of(
        len(data),
        len(serialized_data),
        content_type=content_type,
        content_encoding=ContentEncoding.IDENTITY,
        source_content_length=data.nbytes,
    )

    manifest = Manifest.of(
        entries=ManifestEntryList.of(
            [
                ManifestEntry.of(
                    uri=uri, url=uri, meta=meta, mandatory=True, uuid=manifest_entry_id
                )
            ]
        ),
        author=author,
        uuid=manifest_entry_id,
    )

    delta = Delta.of(
        delta_locator,
        delta_type=delta_type,
        meta=meta,
        properties=properties,
        manifest=manifest,
        previous_stream_position=partition.stream_position,
    )

    params = (uri, serialized_data)
    cur.execute("INSERT OR IGNORE INTO data VALUES (?, ?)", params)

    params = (delta_locator.canonical_string(), "staged_delta", json.dumps(delta))
    cur.execute("INSERT OR IGNORE INTO deltas VALUES (?, ?, ?)", params)

    con.commit()
    return delta


def commit_delta(delta: Delta, *args, **kwargs) -> Delta:
    cur, con = _get_sqlite3_cursor_con(kwargs)

    if not delta.stream_position:
        delta.locator.stream_position = current_time_ms()

    params = (
        delta.locator.canonical_string(),
        delta.partition_locator.canonical_string(),
        json.dumps(delta),
    )

    cur.execute("INSERT OR IGNORE INTO deltas VALUES (?, ?, ?)", params)

    params = (
        delta.partition_locator.canonical_string(),
        json.dumps(delta),
        delta.locator.canonical_string(),
    )
    cur.execute(
        "UPDATE deltas SET partition_locator = ?, value = ? WHERE locator = ?", params
    )

    con.commit()

    return delta


def get_namespace(namespace: str, *args, **kwargs) -> Optional[Namespace]:
    cur, con = _get_sqlite3_cursor_con(kwargs)
    locator = NamespaceLocator.of(namespace)

    res = cur.execute(
        "SELECT * FROM namespaces WHERE locator = ?", (locator.canonical_string(),)
    )
    serialized_result = res.fetchone()

    if serialized_result is None:
        return None

    return Namespace(json.loads(serialized_result[1]))


def namespace_exists(namespace: str, *args, **kwargs) -> bool:
    obj = get_namespace(namespace, *args, **kwargs)

    return obj is not None


def get_table(namespace: str, table_name: str, *args, **kwargs) -> Optional[Table]:
    cur, con = _get_sqlite3_cursor_con(kwargs)
    locator = TableLocator.of(NamespaceLocator.of(namespace), table_name)

    res = cur.execute(
        "SELECT * FROM tables WHERE locator = ?", (locator.canonical_string(),)
    )
    serialized_result = res.fetchone()

    if serialized_result is None:
        return None

    return Table(json.loads(serialized_result[2]))


def table_exists(namespace: str, table_name: str, *args, **kwargs) -> bool:
    obj = get_table(namespace, table_name, *args, **kwargs)

    return obj is not None


def get_table_version(
    namespace: str, table_name: str, table_version: str, *args, **kwargs
) -> Optional[TableVersion]:
    cur, con = _get_sqlite3_cursor_con(kwargs)
    locator = TableVersionLocator.of(
        TableLocator.of(NamespaceLocator.of(namespace), table_name), table_version
    )

    res = cur.execute(
        "SELECT * FROM table_versions WHERE locator = ?", (locator.canonical_string(),)
    )
    serialized_table_version = res.fetchone()

    if serialized_table_version is None:
        return None

    return TableVersion(json.loads(serialized_table_version[2]))


def get_latest_table_version(
    namespace: str, table_name: str, *args, **kwargs
) -> Optional[TableVersion]:
    table_versions = list_table_versions(
        namespace, table_name, *args, **kwargs
    ).all_items()
    if not table_versions:
        return None

    table_versions.sort(reverse=True, key=lambda v: int(v.table_version))
    return table_versions[0]


def get_latest_active_table_version(
    namespace: str, table_name: str, *args, **kwargs
) -> Optional[TableVersion]:

    # This module does not support table version lifecycle state
    return get_latest_table_version(namespace, table_name, *args, **kwargs)


def get_table_version_schema(
    namespace: str,
    table_name: str,
    table_version: Optional[str] = None,
    *args,
    **kwargs,
) -> Optional[Union[pa.Schema, str, bytes]]:
    obj = get_table_version(namespace, table_name, table_version, *args, **kwargs)

    return obj.schema


def table_version_exists(
    namespace: str, table_name: str, table_version: str, *args, **kwargs
) -> bool:
    obj = get_table_version(namespace, table_name, table_version, *args, **kwargs)

    return obj is not None


def get_stream(
    namespace: str,
    table_name: str,
    table_version: Optional[str] = None,
    *args,
    **kwargs,
) -> Optional[Stream]:
    assert not isinstance(table_version, int), f"Passed an integer as the table version"
    obj = get_table_version(namespace, table_name, table_version, *args, **kwargs)

    if obj is None:
        return None

    stream_id = obj.properties.get(STREAM_ID_PROPERTY)
    if stream_id is None:
        return None

    cur, con = _get_sqlite3_cursor_con(kwargs)
    stream_locator = StreamLocator.of(
        obj.locator, stream_id=stream_id, storage_type=STORAGE_TYPE
    )
    res = cur.execute(
        "SELECT * FROM streams WHERE locator = ?", (stream_locator.canonical_string(),)
    )

    serialized_stream = res.fetchone()
    if serialized_stream is None:
        return None

    return Stream(json.loads(serialized_stream[2]))


def get_table_version_column_names(
    namespace: str,
    table_name: str,
    table_version: Optional[str] = None,
    *args,
    **kwargs,
) -> Optional[List[str]]:
    raise NotImplementedError("Fetching column names is not supported")
